"""
Windows specific utility functions, this module should be imported in a try,
except block because it is only applicable on Windows platforms.


Much of what is here was adapted from the following:

    https://stackoverflow.com/a/43233332
    http://stackoverflow.com/questions/29566330
"""

import collections
import ctypes
import logging
import os
from ctypes import wintypes

# pylint: disable=3rd-party-module-not-gated
import ntsecuritycon
import psutil
import win32api
import win32con
import win32process
import win32security
import win32service

import salt.utils.path

# pylint: enable=3rd-party-module-not-gated

# Set up logging
log = logging.getLogger(__name__)

ntdll = ctypes.WinDLL("ntdll")
secur32 = ctypes.WinDLL("secur32")
kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)
advapi32 = ctypes.WinDLL("advapi32", use_last_error=True)
userenv = ctypes.WinDLL("userenv", use_last_error=True)

SYSTEM_SID = "S-1-5-18"
LOCAL_SRV_SID = "S-1-5-19"
NETWORK_SRV_SID = "S-1-5-19"

LOGON_WITH_PROFILE = 0x00000001

WINSTA_ALL = (
    win32con.WINSTA_ACCESSCLIPBOARD
    | win32con.WINSTA_ACCESSGLOBALATOMS
    | win32con.WINSTA_CREATEDESKTOP
    | win32con.WINSTA_ENUMDESKTOPS
    | win32con.WINSTA_ENUMERATE
    | win32con.WINSTA_EXITWINDOWS
    | win32con.WINSTA_READATTRIBUTES
    | win32con.WINSTA_READSCREEN
    | win32con.WINSTA_WRITEATTRIBUTES
    | win32con.DELETE
    | win32con.READ_CONTROL
    | win32con.WRITE_DAC
    | win32con.WRITE_OWNER
)

DESKTOP_ALL = (
    win32con.DESKTOP_CREATEMENU
    | win32con.DESKTOP_CREATEWINDOW
    | win32con.DESKTOP_ENUMERATE
    | win32con.DESKTOP_HOOKCONTROL
    | win32con.DESKTOP_JOURNALPLAYBACK
    | win32con.DESKTOP_JOURNALRECORD
    | win32con.DESKTOP_READOBJECTS
    | win32con.DESKTOP_SWITCHDESKTOP
    | win32con.DESKTOP_WRITEOBJECTS
    | win32con.DELETE
    | win32con.READ_CONTROL
    | win32con.WRITE_DAC
    | win32con.WRITE_OWNER
)

MAX_COMPUTER_NAME_LENGTH = 15

SECURITY_LOGON_TYPE = wintypes.ULONG
Interactive = 2
Network = 3
Batch = 4
Service = 5

LOGON_SUBMIT_TYPE = wintypes.ULONG
PROFILE_BUFFER_TYPE = wintypes.ULONG

MsV1_0InteractiveLogon = 2
MsV1_0Lm20Logon = 3
MsV1_0NetworkLogon = 4
MsV1_0WorkstationUnlockLogon = 7
MsV1_0S4ULogon = 12
MsV1_0NoElevationLogon = 82

KerbInteractiveLogon = 2
KerbWorkstationUnlockLogon = 7
KerbS4ULogon = 12

MSV1_0_S4U_LOGON_FLAG_CHECK_LOGONHOURS = 0x2

KERB_S4U_LOGON_FLAG_CHECK_LOGONHOURS = 0x2
KERB_S4U_LOGON_FLAG_IDENTITY = 0x8

TOKEN_SOURCE_LENGTH = 8

NEGOTIATE_PACKAGE_NAME = b"Negotiate"
MICROSOFT_KERBEROS_NAME = b"Kerberos"
MSV1_0_PACKAGE_NAME = b"MICROSOFT_AUTHENTICATION_PACKAGE_V1_0"

DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000

STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER

TOKEN_ASSIGN_PRIMARY = 0x0001
TOKEN_DUPLICATE = 0x0002
TOKEN_IMPERSONATE = 0x0004
TOKEN_QUERY = 0x0008
TOKEN_QUERY_SOURCE = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS = 0x0040
TOKEN_ADJUST_DEFAULT = 0x0080
TOKEN_ADJUST_SESSIONID = 0x0100

TOKEN_ALL_ACCESS = (
    STANDARD_RIGHTS_REQUIRED
    | TOKEN_ASSIGN_PRIMARY
    | TOKEN_DUPLICATE
    | TOKEN_IMPERSONATE
    | TOKEN_QUERY
    | TOKEN_QUERY_SOURCE
    | TOKEN_ADJUST_PRIVILEGES
    | TOKEN_ADJUST_GROUPS
    | TOKEN_ADJUST_DEFAULT
    | TOKEN_ADJUST_SESSIONID
)

DUPLICATE_CLOSE_SOURCE = 0x00000001
DUPLICATE_SAME_ACCESS = 0x00000002

TOKEN_TYPE = wintypes.ULONG
TokenPrimary = 1
TokenImpersonation = 2

SECURITY_IMPERSONATION_LEVEL = wintypes.ULONG
SecurityAnonymous = 0
SecurityIdentification = 1
SecurityImpersonation = 2
SecurityDelegation = 3


class NTSTATUS(wintypes.LONG):
    def to_error(self):
        return ntdll.RtlNtStatusToDosError(self)

    def __repr__(self):
        name = self.__class__.__name__
        status = wintypes.ULONG.from_buffer(self)
        return f"{name}({status.value})"


PNTSTATUS = ctypes.POINTER(NTSTATUS)


class BOOL(wintypes.BOOL):
    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({bool(self)})"


class HANDLE(wintypes.HANDLE):
    __slots__ = ("closed",)

    def __int__(self):
        return self.value or 0

    def Detach(self):
        if not getattr(self, "closed", False):
            self.closed = True
            value = int(self)
            self.value = None
            return value
        raise ValueError("already closed")

    def Close(self, CloseHandle=kernel32.CloseHandle):
        if self and not getattr(self, "closed", False):
            try:
                CloseHandle(self.Detach())
            except OSError:
                # Suppress the error when there is no handle (WinError 6)
                if ctypes.get_last_error() == 6:
                    pass

    __del__ = Close

    def __repr__(self):
        return f"{self.__class__.__name__}({int(self)})"


class LARGE_INTEGER(wintypes.LARGE_INTEGER):
    # https://msdn.microsoft.com/en-us/library/ff553204
    ntdll.RtlSecondsSince1970ToTime.restype = None
    _unix_epoch = wintypes.LARGE_INTEGER()
    ntdll.RtlSecondsSince1970ToTime(0, ctypes.byref(_unix_epoch))
    _unix_epoch = _unix_epoch.value

    def __int__(self):
        return self.value

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.value})"

    def as_time(self):
        time100ns = self.value - self._unix_epoch
        if time100ns >= 0:
            return time100ns / 1e7
        raise ValueError("value predates the Unix epoch")

    @classmethod
    def from_time(cls, t):
        time100ns = int(t * 10**7)
        return cls(time100ns + cls._unix_epoch)


CHAR = ctypes.c_char
WCHAR = ctypes.c_wchar
PCHAR = ctypes.POINTER(CHAR)
PWCHAR = ctypes.POINTER(WCHAR)


class STRING(ctypes.Structure):
    _fields_ = (
        ("Length", wintypes.USHORT),
        ("MaximumLength", wintypes.USHORT),
        ("Buffer", PCHAR),
    )


LPSTRING = ctypes.POINTER(STRING)


class UNICODE_STRING(ctypes.Structure):
    _fields_ = (
        ("Length", wintypes.USHORT),
        ("MaximumLength", wintypes.USHORT),
        ("Buffer", PWCHAR),
    )


LPUNICODE_STRING = ctypes.POINTER(UNICODE_STRING)


class LUID(ctypes.Structure):
    _fields_ = (
        ("LowPart", wintypes.DWORD),
        ("HighPart", wintypes.LONG),
    )

    def __new__(cls, value=0):
        return cls.from_buffer_copy(ctypes.c_ulonglong(value))

    def __int__(self):
        return ctypes.c_ulonglong.from_buffer(self).value

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({int(self)})"


LPLUID = ctypes.POINTER(LUID)
PSID = wintypes.LPVOID


class SID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("Sid", PSID),
        ("Attributes", wintypes.DWORD),
    )


LPSID_AND_ATTRIBUTES = ctypes.POINTER(SID_AND_ATTRIBUTES)


class TOKEN_GROUPS(ctypes.Structure):
    _fields_ = (
        ("GroupCount", wintypes.DWORD),
        ("Groups", SID_AND_ATTRIBUTES * 1),
    )


LPTOKEN_GROUPS = ctypes.POINTER(TOKEN_GROUPS)


class TOKEN_SOURCE(ctypes.Structure):
    _fields_ = (
        ("SourceName", CHAR * TOKEN_SOURCE_LENGTH),
        ("SourceIdentifier", LUID),
    )

    def __init__(self, SourceName=None, SourceIdentifier=None):
        super().__init__()
        if SourceName is not None:
            if not isinstance(SourceName, bytes):
                SourceName = SourceName.encode("mbcs")
            self.SourceName = SourceName
        if SourceIdentifier is None:
            # pylint: disable=access-member-before-definition
            luid = self.SourceIdentifier
            # pylint: enable=access-member-before-definition
            ntdll.NtAllocateLocallyUniqueId(ctypes.byref(luid))
        else:
            self.SourceIdentifier = SourceIdentifier


LPTOKEN_SOURCE = ctypes.POINTER(TOKEN_SOURCE)
py_source_context = TOKEN_SOURCE(b"PYTHON  ")
py_origin_name = __name__.encode()
py_logon_process_name = f"{py_origin_name}-{os.getpid()}"
SIZE_T = ctypes.c_size_t


class QUOTA_LIMITS(ctypes.Structure):
    _fields_ = (
        ("PagedPoolLimit", SIZE_T),
        ("NonPagedPoolLimit", SIZE_T),
        ("MinimumWorkingSetSize", SIZE_T),
        ("MaximumWorkingSetSize", SIZE_T),
        ("PagefileLimit", SIZE_T),
        ("TimeLimit", wintypes.LARGE_INTEGER),
    )


LPQUOTA_LIMITS = ctypes.POINTER(QUOTA_LIMITS)
LPULONG = ctypes.POINTER(wintypes.ULONG)
LSA_OPERATIONAL_MODE = wintypes.ULONG
LPLSA_OPERATIONAL_MODE = LPULONG
LPHANDLE = ctypes.POINTER(wintypes.HANDLE)
LPLPVOID = ctypes.POINTER(wintypes.LPVOID)
LPDWORD = ctypes.POINTER(wintypes.DWORD)


class ContiguousUnicode(ctypes.Structure):
    # _string_names_: sequence matched to underscore-prefixed fields
    def __init__(self, *args, **kwargs):  # pylint: disable=useless-super-delegation
        super().__init__(*args, **kwargs)

    def _get_unicode_string(self, name):
        wchar_size = ctypes.sizeof(WCHAR)
        s = getattr(self, f"_{name}")
        length = s.Length // wchar_size
        buf = s.Buffer
        if buf:
            return buf[:length]
        return None

    def _set_unicode_buffer(self, values):
        cls = type(self)
        wchar_size = ctypes.sizeof(WCHAR)
        bufsize = (len("\x00".join(values)) + 1) * wchar_size
        ctypes.resize(self, ctypes.sizeof(cls) + bufsize)
        addr = ctypes.addressof(self) + ctypes.sizeof(cls)
        for value in values:
            bufsize = (len(value) + 1) * wchar_size
            ctypes.memmove(addr, value, bufsize)
            addr += bufsize

    def _set_unicode_string(self, name, value):
        values = []
        for n in self._string_names_:
            if n == name:
                values.append(value or "")
            else:
                values.append(getattr(self, n) or "")
        self._set_unicode_buffer(values)

        cls = type(self)
        wchar_size = ctypes.sizeof(WCHAR)
        addr = ctypes.addressof(self) + ctypes.sizeof(cls)
        for n, v in zip(self._string_names_, values):
            ptr = ctypes.cast(addr, PWCHAR)
            ustr = getattr(self, f"_{n}")
            length = ustr.Length = len(v) * wchar_size
            full_length = length + wchar_size
            if (n == name and value is None) or (
                n != name and not (length or ustr.Buffer)
            ):
                ustr.Buffer = None
                ustr.MaximumLength = 0
            else:
                ustr.Buffer = ptr
                ustr.MaximumLength = full_length
            addr += full_length

    def __getattr__(self, name):
        if name not in self._string_names_:
            raise AttributeError
        return self._get_unicode_string(name)

    def __setattr__(self, name, value):
        if name in self._string_names_:
            self._set_unicode_string(name, value)
        else:
            super().__setattr__(name, value)

    @classmethod
    def from_address_copy(cls, address, size=None):
        x = ctypes.Structure.__new__(cls)  # pylint: disable=no-value-for-parameter
        if size is not None:
            ctypes.resize(x, size)
        ctypes.memmove(ctypes.byref(x), address, ctypes.sizeof(x))
        delta = ctypes.addressof(x) - address
        for n in cls._string_names_:
            ustr = getattr(x, f"_{n}")
            addr = ctypes.c_void_p.from_buffer(ustr.Buffer)
            if addr:
                addr.value += delta
        return x


class AuthInfo(ContiguousUnicode):
    # _message_type_: from a logon-submit-type enumeration
    def __init__(self):
        super().__init__()
        self.MessageType = self._message_type_


class MSV1_0_INTERACTIVE_LOGON(AuthInfo):
    _message_type_ = MsV1_0InteractiveLogon
    _string_names_ = "LogonDomainName", "UserName", "Password"

    _fields_ = (
        ("MessageType", LOGON_SUBMIT_TYPE),
        ("_LogonDomainName", UNICODE_STRING),
        ("_UserName", UNICODE_STRING),
        ("_Password", UNICODE_STRING),
    )

    def __init__(self, UserName=None, Password=None, LogonDomainName=None):
        super().__init__()
        if LogonDomainName is not None:
            self.LogonDomainName = LogonDomainName
        if UserName is not None:
            self.UserName = UserName
        if Password is not None:
            self.Password = Password


class S4ULogon(AuthInfo):
    _string_names_ = "UserPrincipalName", "DomainName"

    _fields_ = (
        ("MessageType", LOGON_SUBMIT_TYPE),
        ("Flags", wintypes.ULONG),
        ("_UserPrincipalName", UNICODE_STRING),
        ("_DomainName", UNICODE_STRING),
    )

    def __init__(self, UserPrincipalName=None, DomainName=None, Flags=0):
        super().__init__()
        self.Flags = Flags
        if UserPrincipalName is not None:
            self.UserPrincipalName = UserPrincipalName
        if DomainName is not None:
            self.DomainName = DomainName


class MSV1_0_S4U_LOGON(S4ULogon):
    _message_type_ = MsV1_0S4ULogon


class KERB_S4U_LOGON(S4ULogon):
    _message_type_ = KerbS4ULogon


PMSV1_0_S4U_LOGON = ctypes.POINTER(MSV1_0_S4U_LOGON)
PKERB_S4U_LOGON = ctypes.POINTER(KERB_S4U_LOGON)


class ProfileBuffer(ContiguousUnicode):
    # _message_type_
    def __init__(self):
        super().__init__()
        self.MessageType = self._message_type_


class MSV1_0_INTERACTIVE_PROFILE(ProfileBuffer):
    _message_type_ = MsV1_0InteractiveLogon
    _string_names_ = (
        "LogonScript",
        "HomeDirectory",
        "FullName",
        "ProfilePath",
        "HomeDirectoryDrive",
        "LogonServer",
    )
    _fields_ = (
        ("MessageType", PROFILE_BUFFER_TYPE),
        ("LogonCount", wintypes.USHORT),
        ("BadPasswordCount", wintypes.USHORT),
        ("LogonTime", LARGE_INTEGER),
        ("LogoffTime", LARGE_INTEGER),
        ("KickOffTime", LARGE_INTEGER),
        ("PasswordLastSet", LARGE_INTEGER),
        ("PasswordCanChange", LARGE_INTEGER),
        ("PasswordMustChange", LARGE_INTEGER),
        ("_LogonScript", UNICODE_STRING),
        ("_HomeDirectory", UNICODE_STRING),
        ("_FullName", UNICODE_STRING),
        ("_ProfilePath", UNICODE_STRING),
        ("_HomeDirectoryDrive", UNICODE_STRING),
        ("_LogonServer", UNICODE_STRING),
        ("UserFlags", wintypes.ULONG),
    )


def _check_status(result, func, args):
    if result.value < 0:
        raise ctypes.WinError(result.to_error())
    return args


def _check_bool(result, func, args):
    if not result:
        raise ctypes.WinError(ctypes.get_last_error())
    return args


INVALID_HANDLE_VALUE = wintypes.HANDLE(-1).value
INVALID_DWORD_VALUE = wintypes.DWORD(-1).value  # ~WinAPI
INFINITE = INVALID_DWORD_VALUE
STD_INPUT_HANDLE = wintypes.DWORD(-10).value
STD_OUTPUT_HANDLE = wintypes.DWORD(-11).value
STD_ERROR_HANDLE = wintypes.DWORD(-12).value


class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("nLength", wintypes.DWORD),
        ("lpSecurityDescriptor", wintypes.LPVOID),
        ("bInheritHandle", wintypes.BOOL),
    )

    def __init__(self, **kwds):
        self.nLength = ctypes.sizeof(self)
        super().__init__(**kwds)


LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)
LPBYTE = ctypes.POINTER(wintypes.BYTE)
LPHANDLE = PHANDLE = ctypes.POINTER(ctypes.c_void_p)
LPDWORD = ctypes.POINTER(ctypes.c_ulong)


class STARTUPINFO(ctypes.Structure):
    """https://msdn.microsoft.com/en-us/library/ms686331"""

    _fields_ = (
        ("cb", wintypes.DWORD),
        ("lpReserved", wintypes.LPWSTR),
        ("lpDesktop", wintypes.LPWSTR),
        ("lpTitle", wintypes.LPWSTR),
        ("dwX", wintypes.DWORD),
        ("dwY", wintypes.DWORD),
        ("dwXSize", wintypes.DWORD),
        ("dwYSize", wintypes.DWORD),
        ("dwXCountChars", wintypes.DWORD),
        ("dwYCountChars", wintypes.DWORD),
        ("dwFillAttribute", wintypes.DWORD),
        ("dwFlags", wintypes.DWORD),
        ("wShowWindow", wintypes.WORD),
        ("cbReserved2", wintypes.WORD),
        ("lpReserved2", LPBYTE),
        ("hStdInput", wintypes.HANDLE),
        ("hStdOutput", wintypes.HANDLE),
        ("hStdError", wintypes.HANDLE),
    )

    def __init__(self, **kwds):
        self.cb = ctypes.sizeof(self)
        super().__init__(**kwds)


LPSTARTUPINFO = ctypes.POINTER(STARTUPINFO)


class PROC_THREAD_ATTRIBUTE_LIST(ctypes.Structure):
    pass


PPROC_THREAD_ATTRIBUTE_LIST = ctypes.POINTER(PROC_THREAD_ATTRIBUTE_LIST)


class STARTUPINFOEX(STARTUPINFO):
    _fields_ = (("lpAttributeList", PPROC_THREAD_ATTRIBUTE_LIST),)


LPSTARTUPINFOEX = ctypes.POINTER(STARTUPINFOEX)


class PROCESS_INFORMATION(ctypes.Structure):
    """https://msdn.microsoft.com/en-us/library/ms684873"""

    _fields_ = (
        ("hProcess", wintypes.HANDLE),
        ("hThread", wintypes.HANDLE),
        ("dwProcessId", wintypes.DWORD),
        ("dwThreadId", wintypes.DWORD),
    )


LPPROCESS_INFORMATION = ctypes.POINTER(PROCESS_INFORMATION)


class HANDLE_IHV(wintypes.HANDLE):
    pass


def errcheck_ihv(result, func, args):
    if result.value == INVALID_HANDLE_VALUE:
        raise ctypes.WinError(ctypes.get_last_error())
    return result.value


class DWORD_IDV(wintypes.DWORD):
    pass


def errcheck_idv(result, func, args):
    if result.value == INVALID_DWORD_VALUE:
        raise ctypes.WinError(ctypes.get_last_error())
    return result.value


def errcheck_bool(result, func, args):
    if not result:
        raise ctypes.WinError(ctypes.get_last_error())
    return args


def _win(func, restype, *argtypes):
    func.restype = restype
    func.argtypes = argtypes
    if issubclass(restype, NTSTATUS):
        func.errcheck = _check_status
    elif issubclass(restype, BOOL):
        func.errcheck = _check_bool
    elif issubclass(restype, HANDLE_IHV):
        func.errcheck = errcheck_ihv
    elif issubclass(restype, DWORD_IDV):
        func.errcheck = errcheck_idv
    else:
        func.errcheck = errcheck_bool


# https://msdn.microsoft.com/en-us/library/ms683231
_win(kernel32.GetStdHandle, HANDLE_IHV, wintypes.DWORD)  # _In_ nStdHandle


# https://msdn.microsoft.com/en-us/library/ms724211
_win(kernel32.CloseHandle, wintypes.BOOL, wintypes.HANDLE)  # _In_ hObject


# https://msdn.microsoft.com/en-us/library/ms724935
_win(
    kernel32.SetHandleInformation,
    wintypes.BOOL,
    wintypes.HANDLE,  # _In_ hObject
    wintypes.DWORD,  # _In_ dwMask
    wintypes.DWORD,
)  # _In_ dwFlags


# https://msdn.microsoft.com/en-us/library/ms724251
_win(
    kernel32.DuplicateHandle,
    wintypes.BOOL,
    wintypes.HANDLE,  # _In_  hSourceProcessHandle,
    wintypes.HANDLE,  # _In_  hSourceHandle,
    wintypes.HANDLE,  # _In_  hTargetProcessHandle,
    LPHANDLE,  # _Out_ lpTargetHandle,
    wintypes.DWORD,  # _In_  dwDesiredAccess,
    wintypes.BOOL,  # _In_  bInheritHandle,
    wintypes.DWORD,
)  # _In_  dwOptions


# https://msdn.microsoft.com/en-us/library/ms683179
_win(kernel32.GetCurrentProcess, wintypes.HANDLE)


# https://msdn.microsoft.com/en-us/library/ms683189
_win(
    kernel32.GetExitCodeProcess,
    wintypes.BOOL,
    wintypes.HANDLE,  # _In_  hProcess,
    LPDWORD,
)  # _Out_ lpExitCode


# https://msdn.microsoft.com/en-us/library/aa365152
_win(
    kernel32.CreatePipe,
    wintypes.BOOL,
    PHANDLE,  # _Out_    hReadPipe,
    PHANDLE,  # _Out_    hWritePipe,
    LPSECURITY_ATTRIBUTES,  # _In_opt_ lpPipeAttributes,
    wintypes.DWORD,
)  # _In_     nSize


# https://msdn.microsoft.com/en-us/library/ms682431
# _win(advapi32.CreateProcessWithTokenW, wintypes.BOOL,
#    PHANDLE,       # _In_        lpUsername
#    wintypes.DWORD,         # _In_        dwLogonFlags
#    wintypes.LPCWSTR,       # _In_opt_    lpApplicationName
#    wintypes.LPWSTR,        # _Inout_opt_ lpCommandLine
#    wintypes.DWORD,         # _In_        dwCreationFlags
#    wintypes.LPVOID,        # _In_opt_     lpEnvironment
#    wintypes.LPCWSTR,       # _In_opt_    lpCurrentDirectory
#    LPSTARTUPINFO,          # _In_        lpStartupInfo
#    LPPROCESS_INFORMATION)  # _Out_       lpProcessInformation


# https://msdn.microsoft.com/en-us/library/ms682431
_win(
    advapi32.CreateProcessWithLogonW,
    wintypes.BOOL,
    wintypes.LPCWSTR,  # _In_        lpUsername
    wintypes.LPCWSTR,  # _In_opt_    lpDomain
    wintypes.LPCWSTR,  # _In_        lpPassword
    wintypes.DWORD,  # _In_        dwLogonFlags
    wintypes.LPCWSTR,  # _In_opt_    lpApplicationName
    wintypes.LPWSTR,  # _Inout_opt_ lpCommandLine
    wintypes.DWORD,  # _In_        dwCreationFlags
    wintypes.LPCWSTR,  # _In_opt_    lpEnvironment
    wintypes.LPCWSTR,  # _In_opt_    lpCurrentDirectory
    LPSTARTUPINFO,  # _In_        lpStartupInfo
    LPPROCESS_INFORMATION,
)  # _Out_       lpProcessInformation


# https://msdn.microsoft.com/en-us/library/ms683179
_win(kernel32.GetCurrentProcess, wintypes.HANDLE)


# https://msdn.microsoft.com/en-us/library/ms724251
_win(
    kernel32.DuplicateHandle,
    BOOL,
    wintypes.HANDLE,  # _In_  hSourceProcessHandle
    wintypes.HANDLE,  # _In_  hSourceHandle
    wintypes.HANDLE,  # _In_  hTargetProcessHandle
    LPHANDLE,  # _Out_ lpTargetHandle
    wintypes.DWORD,  # _In_  dwDesiredAccess
    wintypes.BOOL,  # _In_  bInheritHandle
    wintypes.DWORD,
)  # _In_  dwOptions


# https://msdn.microsoft.com/en-us/library/ms724295
_win(
    kernel32.GetComputerNameW, BOOL, wintypes.LPWSTR, LPDWORD  # _Out_   lpBuffer
)  # _Inout_ lpnSize


# https://msdn.microsoft.com/en-us/library/aa379295
_win(
    advapi32.OpenProcessToken,
    BOOL,
    wintypes.HANDLE,  # _In_  ProcessHandle
    wintypes.DWORD,  # _In_  DesiredAccess
    LPHANDLE,
)  # _Out_ TokenHandle


# https://msdn.microsoft.com/en-us/library/aa446617
_win(
    advapi32.DuplicateTokenEx,
    BOOL,
    wintypes.HANDLE,  # _In_     hExistingToken
    wintypes.DWORD,  # _In_     dwDesiredAccess
    LPSECURITY_ATTRIBUTES,  # _In_opt_ lpTokenAttributes
    SECURITY_IMPERSONATION_LEVEL,  # _In_     ImpersonationLevel
    TOKEN_TYPE,  # _In_     TokenType
    LPHANDLE,
)  # _Out_    phNewToken


# https://msdn.microsoft.com/en-us/library/ff566415
_win(ntdll.NtAllocateLocallyUniqueId, NTSTATUS, LPLUID)  # _Out_ LUID


# https://msdn.microsoft.com/en-us/library/aa378279
_win(
    secur32.LsaFreeReturnBuffer,
    NTSTATUS,
    wintypes.LPVOID,
)  # _In_ Buffer


# https://msdn.microsoft.com/en-us/library/aa378265
_win(
    secur32.LsaConnectUntrusted,
    NTSTATUS,
    LPHANDLE,
)  # _Out_ LsaHandle


# https://msdn.microsoft.com/en-us/library/aa378318
_win(
    secur32.LsaRegisterLogonProcess,
    NTSTATUS,
    LPSTRING,  # _In_  LogonProcessName
    LPHANDLE,  # _Out_ LsaHandle
    LPLSA_OPERATIONAL_MODE,
)  # _Out_ SecurityMode


# https://msdn.microsoft.com/en-us/library/aa378269
_win(secur32.LsaDeregisterLogonProcess, NTSTATUS, wintypes.HANDLE)  # _In_ LsaHandle


# https://msdn.microsoft.com/en-us/library/aa378297
_win(
    secur32.LsaLookupAuthenticationPackage,
    NTSTATUS,
    wintypes.HANDLE,  # _In_  LsaHandle
    LPSTRING,  # _In_  PackageName
    LPULONG,
)  # _Out_ AuthenticationPackage


# https://msdn.microsoft.com/en-us/library/aa378292
_win(
    secur32.LsaLogonUser,
    NTSTATUS,
    wintypes.HANDLE,  # _In_     LsaHandle
    LPSTRING,  # _In_     OriginName
    SECURITY_LOGON_TYPE,  # _In_     LogonType
    wintypes.ULONG,  # _In_     AuthenticationPackage
    wintypes.LPVOID,  # _In_     AuthenticationInformation
    wintypes.ULONG,  # _In_     AuthenticationInformationLength
    LPTOKEN_GROUPS,  # _In_opt_ LocalGroups
    LPTOKEN_SOURCE,  # _In_     SourceContext
    LPLPVOID,  # _Out_    ProfileBuffer
    LPULONG,  # _Out_    ProfileBufferLength
    LPLUID,  # _Out_    LogonId
    LPHANDLE,  # _Out_    Token
    LPQUOTA_LIMITS,  # _Out_    Quotas
    PNTSTATUS,
)  # _Out_    SubStatus


def duplicate_token(
    source_token=None,
    access=TOKEN_ALL_ACCESS,
    impersonation_level=SecurityImpersonation,
    token_type=TokenPrimary,
    attributes=None,
):
    close_source = False
    if source_token is None:
        close_source = True
        source_token = HANDLE()
        advapi32.OpenProcessToken(
            kernel32.GetCurrentProcess(), TOKEN_ALL_ACCESS, ctypes.byref(source_token)
        )
    token = HANDLE()
    try:
        advapi32.DuplicateTokenEx(
            source_token,
            access,
            attributes,
            impersonation_level,
            token_type,
            ctypes.byref(token),
        )
    finally:
        if close_source:
            source_token.Close()
    return token


def lsa_connect_untrusted():
    handle = wintypes.HANDLE()
    secur32.LsaConnectUntrusted(ctypes.byref(handle))
    return handle.value


def lsa_register_logon_process(logon_process_name):
    if not isinstance(logon_process_name, bytes):
        logon_process_name = logon_process_name.encode("mbcs")
    logon_process_name = logon_process_name[:127]
    buf = ctypes.create_string_buffer(logon_process_name, 128)
    name = STRING(len(logon_process_name), len(buf), buf)
    handle = wintypes.HANDLE()
    mode = LSA_OPERATIONAL_MODE()
    secur32.LsaRegisterLogonProcess(
        ctypes.byref(name), ctypes.byref(handle), ctypes.byref(mode)
    )
    return handle.value


def lsa_lookup_authentication_package(lsa_handle, package_name):
    if not isinstance(package_name, bytes):
        package_name = package_name.encode("mbcs")
    package_name = package_name[:127]
    buf = ctypes.create_string_buffer(package_name)
    name = STRING(len(package_name), len(buf), buf)
    package = wintypes.ULONG()
    secur32.LsaLookupAuthenticationPackage(
        lsa_handle, ctypes.byref(name), ctypes.byref(package)
    )
    return package.value


LOGONINFO = collections.namedtuple(
    "LOGONINFO", ("Token", "LogonId", "Profile", "Quotas")
)


def lsa_logon_user(
    auth_info,
    local_groups=None,
    origin_name=py_origin_name,
    source_context=None,
    auth_package=None,
    logon_type=None,
    lsa_handle=None,
):
    if local_groups is None:
        plocal_groups = LPTOKEN_GROUPS()
    else:
        plocal_groups = ctypes.byref(local_groups)
    if source_context is None:
        source_context = py_source_context
    if not isinstance(origin_name, bytes):
        origin_name = origin_name.encode("mbcs")
    buf = ctypes.create_string_buffer(origin_name)
    origin_name = STRING(len(origin_name), len(buf), buf)
    if auth_package is None:
        if isinstance(auth_info, MSV1_0_S4U_LOGON):
            auth_package = NEGOTIATE_PACKAGE_NAME
        elif isinstance(auth_info, KERB_S4U_LOGON):
            auth_package = MICROSOFT_KERBEROS_NAME
        else:
            auth_package = MSV1_0_PACKAGE_NAME
    if logon_type is None:
        if isinstance(auth_info, S4ULogon):
            logon_type = win32con.LOGON32_LOGON_NETWORK
        else:
            logon_type = Interactive
    profile_buffer = wintypes.LPVOID()
    profile_buffer_length = wintypes.ULONG()
    profile = None
    logonid = LUID()
    htoken = HANDLE()
    quotas = QUOTA_LIMITS()
    substatus = NTSTATUS()
    deregister = False
    if lsa_handle is None:
        lsa_handle = lsa_connect_untrusted()
        deregister = True
    try:
        if isinstance(auth_package, (str, bytes)):
            auth_package = lsa_lookup_authentication_package(lsa_handle, auth_package)
        try:
            secur32.LsaLogonUser(
                lsa_handle,
                ctypes.byref(origin_name),
                logon_type,
                auth_package,
                ctypes.byref(auth_info),
                ctypes.sizeof(auth_info),
                plocal_groups,
                ctypes.byref(source_context),
                ctypes.byref(profile_buffer),
                ctypes.byref(profile_buffer_length),
                ctypes.byref(logonid),
                ctypes.byref(htoken),
                ctypes.byref(quotas),
                ctypes.byref(substatus),
            )
        except OSError:
            if substatus.value:
                raise ctypes.WinError(substatus.to_error())
            raise
        finally:
            if profile_buffer:
                address = profile_buffer.value
                buftype = PROFILE_BUFFER_TYPE.from_address(address).value
                if buftype == MsV1_0InteractiveLogon:
                    profile = MSV1_0_INTERACTIVE_PROFILE.from_address_copy(
                        address, profile_buffer_length.value
                    )
                secur32.LsaFreeReturnBuffer(address)
    finally:
        if deregister:
            secur32.LsaDeregisterLogonProcess(lsa_handle)
    return LOGONINFO(htoken, logonid, profile, quotas)


def logon_msv1(
    name,
    password,
    domain=None,
    local_groups=None,
    origin_name=py_origin_name,
    source_context=None,
):
    return lsa_logon_user(
        MSV1_0_INTERACTIVE_LOGON(name, password, domain),
        local_groups,
        origin_name,
        source_context,
    )


def logon_msv1_s4u(
    name, local_groups=None, origin_name=py_origin_name, source_context=None
):
    domain = ctypes.create_unicode_buffer(MAX_COMPUTER_NAME_LENGTH + 1)
    length = wintypes.DWORD(len(domain))
    kernel32.GetComputerNameW(domain, ctypes.byref(length))
    return lsa_logon_user(
        MSV1_0_S4U_LOGON(name, domain.value), local_groups, origin_name, source_context
    )


def logon_kerb_s4u(
    name,
    realm=None,
    local_groups=None,
    origin_name=py_origin_name,
    source_context=None,
    logon_process_name=py_logon_process_name,
):
    lsa_handle = lsa_register_logon_process(logon_process_name)
    try:
        return lsa_logon_user(
            KERB_S4U_LOGON(name, realm),
            local_groups,
            origin_name,
            source_context,
            lsa_handle=lsa_handle,
        )
    finally:
        secur32.LsaDeregisterLogonProcess(lsa_handle)


def DuplicateHandle(
    hsrc=kernel32.GetCurrentProcess(),
    srchandle=kernel32.GetCurrentProcess(),
    htgt=kernel32.GetCurrentProcess(),
    access=0,
    inherit=False,
    options=win32con.DUPLICATE_SAME_ACCESS,
):
    tgthandle = wintypes.HANDLE()
    kernel32.DuplicateHandle(
        hsrc, srchandle, htgt, ctypes.byref(tgthandle), access, inherit, options
    )
    return tgthandle.value


def CreatePipe(inherit_read=False, inherit_write=False):
    read, write = wintypes.HANDLE(), wintypes.HANDLE()
    kernel32.CreatePipe(ctypes.byref(read), ctypes.byref(write), None, 0)
    if inherit_read:
        kernel32.SetHandleInformation(
            read, win32con.HANDLE_FLAG_INHERIT, win32con.HANDLE_FLAG_INHERIT
        )
    if inherit_write:
        kernel32.SetHandleInformation(
            write, win32con.HANDLE_FLAG_INHERIT, win32con.HANDLE_FLAG_INHERIT
        )
    return read.value, write.value


def set_user_perm(obj, perm, sid):
    """
    Set an object permission for the given user sid
    """
    info = (
        win32security.OWNER_SECURITY_INFORMATION
        | win32security.GROUP_SECURITY_INFORMATION
        | win32security.DACL_SECURITY_INFORMATION
    )
    sd = win32security.GetUserObjectSecurity(obj, info)
    dacl = sd.GetSecurityDescriptorDacl()
    ace_cnt = dacl.GetAceCount()
    found = False
    for idx in range(0, ace_cnt):
        (aceType, aceFlags), ace_mask, ace_sid = dacl.GetAce(idx)
        ace_exists = (
            aceType == ntsecuritycon.ACCESS_ALLOWED_ACE_TYPE
            and ace_mask == perm
            and ace_sid == sid
        )
        if ace_exists:
            # If the ace already exists, do nothing
            break
    else:
        dacl.AddAccessAllowedAce(dacl.GetAclRevision(), perm, sid)
        sd.SetSecurityDescriptorDacl(1, dacl, 0)
        win32security.SetUserObjectSecurity(obj, info, sd)


def grant_winsta_and_desktop(th):
    """
    Grant the token's user access to the current process's window station and
    desktop.
    """
    current_sid = win32security.GetTokenInformation(th, win32security.TokenUser)[0]
    # Add permissions for the sid to the current windows station and thread id.
    # This prevents windows error 0xC0000142.
    winsta = win32process.GetProcessWindowStation()
    set_user_perm(winsta, WINSTA_ALL, current_sid)
    desktop = win32service.GetThreadDesktop(win32api.GetCurrentThreadId())
    set_user_perm(desktop, DESKTOP_ALL, current_sid)


def environment_string(env):
    senv = ""
    for k, v in env.items():
        senv += k + "=" + v + "\0"
    senv += "\0"
    return ctypes.create_unicode_buffer(senv)


def CreateProcessWithTokenW(
    token,
    logonflags=0,
    applicationname=None,
    commandline=None,
    creationflags=0,
    environment=None,
    currentdirectory=None,
    startupinfo=None,
):
    creationflags |= win32con.CREATE_UNICODE_ENVIRONMENT
    if commandline is not None:
        commandline = ctypes.create_unicode_buffer(commandline)
    if startupinfo is None:
        startupinfo = STARTUPINFO()
    if currentdirectory is not None:
        currentdirectory = ctypes.create_unicode_buffer(currentdirectory)
    if environment is not None:
        environment = ctypes.pointer(environment_string(environment))
    process_info = PROCESS_INFORMATION()
    ret = advapi32.CreateProcessWithTokenW(
        token,
        logonflags,
        applicationname,
        commandline,
        creationflags,
        environment,
        currentdirectory,
        ctypes.byref(startupinfo),
        ctypes.byref(process_info),
    )
    if ret == 0:
        winerr = win32api.GetLastError()
        exc = OSError(win32api.FormatMessage(winerr))
        exc.winerror = winerr
        raise exc
    return process_info


def enumerate_tokens(sid=None, session_id=None, privs=None):
    """
    Enumerate tokens from any existing processes that can be accessed.
    Optionally filter by sid.
    """
    for p in psutil.process_iter():
        if p.pid == 0:
            continue
        try:
            ph = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, 0, p.pid)
        except win32api.error as exc:
            if exc.winerror == 5:
                log.debug("Unable to OpenProcess pid=%d name=%s", p.pid, p.name())
                continue
            raise exc
        try:
            access = (
                win32security.TOKEN_DUPLICATE
                | win32security.TOKEN_QUERY
                | win32security.TOKEN_IMPERSONATE
                | win32security.TOKEN_ASSIGN_PRIMARY
            )
            th = win32security.OpenProcessToken(ph, access)
        except Exception as exc:  # pylint: disable=broad-except
            log.debug(
                "OpenProcessToken failed pid=%d name=%s user%s",
                p.pid,
                p.name(),
                p.username(),
            )
            continue
        try:
            process_sid = win32security.GetTokenInformation(
                th, win32security.TokenUser
            )[0]
        except Exception as exc:  # pylint: disable=broad-except
            log.exception(
                "GetTokenInformation pid=%d name=%s user%s",
                p.pid,
                p.name(),
                p.username(),
            )
            continue

        proc_sid = win32security.ConvertSidToStringSid(process_sid)
        if sid and sid != proc_sid:
            log.debug("Token for pid does not match user sid: %s", sid)
            continue

        if (
            session_id
            and win32security.GetTokenInformation(th, win32security.TokenSessionId)
            != session_id
        ):
            continue

        def has_priv(tok, priv):
            luid = win32security.LookupPrivilegeValue(None, priv)
            for priv_luid, flags in win32security.GetTokenInformation(
                tok, win32security.TokenPrivileges
            ):
                if priv_luid == luid:
                    return True
            return False

        if privs:
            has_all = True
            for name in privs:
                if not has_priv(th, name):
                    has_all = False
            if not has_all:
                continue
        yield dup_token(th)


def impersonate_sid(sid, session_id=None, privs=None):
    """
    Find an existing process token for the given sid and impersonate the token.
    """
    for tok in enumerate_tokens(sid, session_id, privs):
        tok = dup_token(tok)
        elevate_token(tok)
        if win32security.ImpersonateLoggedOnUser(tok) == 0:
            raise OSError("Impersonation failure")
        return tok
    raise OSError("Impersonation failure")


def dup_token(th):
    """
    duplicate the access token
    """
    # TODO: is `duplicate_token` the same?
    sec_attr = win32security.SECURITY_ATTRIBUTES()
    sec_attr.bInheritHandle = True
    return win32security.DuplicateTokenEx(
        th,
        win32security.SecurityImpersonation,
        win32con.MAXIMUM_ALLOWED,
        win32security.TokenPrimary,
        sec_attr,
    )


def elevate_token(th):
    """
    Set all token privileges to enabled
    """
    # Get list of privileges this token contains
    privileges = win32security.GetTokenInformation(th, win32security.TokenPrivileges)

    # Create a set of all privileges to be enabled
    enable_privs = set()
    for luid, flags in privileges:
        enable_privs.add((luid, win32con.SE_PRIVILEGE_ENABLED))

    # Enable the privileges
    if win32security.AdjustTokenPrivileges(th, 0, enable_privs) == 0:
        raise OSError(win32api.FormatMessage(win32api.GetLastError()))


def make_inheritable(token):
    """Create an inheritable handle"""
    return win32api.DuplicateHandle(
        win32api.GetCurrentProcess(),
        token,
        win32api.GetCurrentProcess(),
        0,
        1,
        win32con.DUPLICATE_SAME_ACCESS,
    )


def CreateProcessWithLogonW(
    username=None,
    domain=None,
    password=None,
    logonflags=0,
    applicationname=None,
    commandline=None,
    creationflags=0,
    environment=None,
    currentdirectory=None,
    startupinfo=None,
):
    creationflags |= win32con.CREATE_UNICODE_ENVIRONMENT
    if commandline is not None:
        commandline = ctypes.create_unicode_buffer(commandline)
    if startupinfo is None:
        startupinfo = STARTUPINFO()
    if environment is not None:
        environment = ctypes.pointer(environment_string(environment))
    process_info = PROCESS_INFORMATION()
    advapi32.CreateProcessWithLogonW(
        username,
        domain,
        password,
        logonflags,
        applicationname,
        commandline,
        creationflags,
        environment,
        currentdirectory,
        ctypes.byref(startupinfo),
        ctypes.byref(process_info),
    )
    return process_info


def prepend_cmd(cmd):
    # Some commands are only available when run from a cmd shell. These are
    # built-in commands such as echo. So, let's check for the binary in the
    # path. If it does not exist, let's assume it's a built-in and requires us
    # to run it in a cmd prompt.
    if not isinstance(cmd, (list, tuple)):
        cmd = cmd.split()

    # Let's try to figure out what the fist command is so we can check for
    # builtin commands such as echo
    first_cmd = cmd[0].split("\\")[-1].strip("\"'")

    cmd = " ".join(cmd)

    # Known builtin cmd commands
    known_builtins = [
        "assoc",
        "break",
        "call",
        "cd",
        "chdir",
        "cls",
        "color",
        "copy",
        "date",
        "del",
        "dir",
        "echo",
        "endlocal",
        "erase",
        "exit",
        "for",
        "ftype",
        "goto",
        "if",
        "md",
        "mklink",
        "move",
        "path",
        "pause",
        "popd",
        "prompt",
        "pushd",
        "rd",
        "rem",
        "ren",
        "rmdir",
        "set",
        "setlocal",
        "shift",
        "start",
        "time",
        "title",
        "type",
        "ver",
        "verify",
        "vol",
        "::",
    ]

    # If the first command is one of the known builtin commands or if it is a
    # binary that can't be found, we'll prepend cmd /c
    if first_cmd in known_builtins or salt.utils.path.which(first_cmd) is None:
        cmd = f'cmd /c "{cmd}"'

    # There are a few other things we need to check for that require cmd. If the
    # cmd contains any of the following, we'll need to make sure it runs in cmd.
    # We'll add to this list as more things are discovered.
    check = ["&&", "||"]
    if not cmd.startswith("cmd") and any(chk in cmd for chk in check):
        cmd = f'cmd /c "{cmd}"'

    return cmd



import pytest
from saltfactories.utils.functional import StateResult

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


def test_dateutils_strftime(salt_master, salt_ssh_cli):
    """
    test jinja filter datautils.strftime
    """
    sls_contents = """
    {% set result = none | strftime('%Y-%m-%d') %}
    test:
      module.run:
        - name: test.echo
        - text: {{ result }}
    """
    with salt_master.state_tree.base.temp_file("dateutils.sls", sls_contents):
        ret = salt_ssh_cli.run("state.sls", "dateutils")
        assert ret.returncode == 0
        staterun = StateResult(ret.data)
        assert staterun.result is True
        assert staterun.changes
        assert "ret" in staterun.changes



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def pkg_def_contents(state_tree):
    return r"""
    my-software:
      '1.0.1':
        full_name: 'My Software'
        installer: 'C:\files\mysoftware101.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\mysoftware101.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
      '1.0.2':
        full_name: 'My Software'
        installer: 'C:\files\mysoftware102.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\mysoftware102.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
    your-software:
      '1.0.0':
        full_name: 'Your Software'
        installer: 'C:\files\yoursoftware101.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\yoursoftware101.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
    """


@pytest.fixture(scope="module")
def pkg(modules):
    yield modules.pkg


@pytest.fixture(scope="function")
def repo(pkg, state_tree, pkg_def_contents):
    assert len(pkg.get_package_info("my-software")) == 0
    repo_dir = state_tree / "winrepo_ng"
    with pytest.helpers.temp_file("my-software.sls", pkg_def_contents, repo_dir):
        pkg.refresh_db()


def test_refresh_db(pkg, repo):
    assert len(pkg.get_package_info("my-software")) == 2
    assert len(pkg.get_package_info("your-software")) == 1


@pytest.mark.parametrize(
    "as_dict, reverse, expected",
    [
        (False, False, ["1.0.1", "1.0.2"]),
        (False, True, ["1.0.2", "1.0.1"]),
        (True, False, {"my-software": ["1.0.1", "1.0.2"]}),
        (True, True, {"my-software": ["1.0.2", "1.0.1"]}),
    ],
)
def test_list_available(pkg, repo, as_dict, reverse, expected):
    result = pkg.list_available(
        "my-software", return_dict_always=as_dict, reverse_sort=reverse
    )
    assert result == expected


@pytest.mark.parametrize(
    "pkg_name, expected",
    [
        ("my-software", {"my-software": ["1.0.2", "1.0.1"]}),
        ("my-software=1.0.1", {"my-software": ["1.0.1"]}),
        ("my-soft*", {"my-software": ["1.0.2", "1.0.1"]}),
        ("your-software", {"your-software": ["1.0.0"]}),
        (None, {"my-software": ["1.0.2", "1.0.1"], "your-software": ["1.0.0"]}),
    ],
)
def test_list_repo_pkgs(pkg, repo, pkg_name, expected):
    if pkg_name is None:
        result = pkg.list_repo_pkgs()
    else:
        result = pkg.list_repo_pkgs(pkg_name)
    assert result == expected



import pytest

import salt.utils.files

pytestmark = [
    pytest.mark.slow_test,
]


def test_issue_54765_salt(tmp_path, salt_cli, salt_minion):
    file_path = str(tmp_path / "issue-54765")
    ret = salt_cli.run(
        "state.sls",
        mods="issue-54765",
        pillar={"file_path": file_path},
        minion_tgt=salt_minion.id,
    ).data
    key = f"file_|-issue-54765_|-{file_path}_|-managed"
    assert key in ret
    assert ret[key]["result"] is True
    with salt.utils.files.fopen(file_path, "r") as fp:
        assert fp.read().strip() == "bar"


def test_issue_54765_call(tmp_path, salt_call_cli):
    file_path = str(tmp_path / "issue-54765")
    ret = salt_call_cli.run(
        "--local",
        "state.apply",
        "issue-54765",
        pillar=f"{{'file_path': '{file_path}'}}",
    )
    key = f"file_|-issue-54765_|-{file_path}_|-managed"
    assert ret.data[key]["result"] is True
    with salt.utils.files.fopen(file_path, "r") as fp:
        assert fp.read().strip() == "bar"



"""
Integration tests for the jinja includes in states
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.mark.slow_test
def test_issue_64111(salt_master, salt_minion, salt_call_cli):
    # This needs to be an integration test. A functional test does not trigger
    # the issue fixed.

    macros_jinja = """
    {% macro a_jinja_macro(arg) -%}
    {{ arg }}
    {%- endmacro %}
    """

    init_sls = """
    include:
      - common.file1
    """

    file1_sls = """
    {% from 'common/macros.jinja' import a_jinja_macro with context %}

    a state id:
      cmd.run:
        - name: echo {{ a_jinja_macro("hello world") }}
    """
    tf = salt_master.state_tree.base.temp_file

    with tf("common/macros.jinja", macros_jinja):
        with tf("common/init.sls", init_sls):
            with tf("common/file1.sls", file1_sls):
                ret = salt_call_cli.run("state.apply", "common")
                assert ret.returncode == 0


@pytest.mark.slow_test
def test_issue_65080(salt_master, salt_minion, salt_call_cli):
    """
    Test scenario when a state includes another state that only includes a third state
    """

    only_include_sls = """
include:
  - includetest.another-include
    """

    another_include_sls = """
/tmp/from-test-file.txt:
  file.managed:
    - contents: Hello from test-file.sls
    """

    init_sls = """
include:
  - includetest.only-include

/tmp/from-init.txt:
  file.managed:
    - contents: Hello from init.sls
    - require:
      - sls: includetest.only-include
    """

    tf = salt_master.state_tree.base.temp_file

    with tf("includetest/init.sls", init_sls):
        with tf("includetest/another-include.sls", another_include_sls):
            with tf("includetest/only-include.sls", only_include_sls):
                ret = salt_call_cli.run("state.apply", "includetest")
                assert ret.returncode == 0

                ret = salt_call_cli.run("state.show_low_sls", "includetest")
                assert "__sls_included_from__" in ret.data[0]
                assert (
                    "includetest.only-include" in ret.data[0]["__sls_included_from__"]
                )


@pytest.mark.slow_test
def test_issue_65080_multiple_includes(salt_master, salt_minion, salt_call_cli):
    """
    Test scenario when a state includes another state that only includes a third state
    """

    only_include_one_sls = """
include:
  - includetest.include-one
    """

    only_include_two_sls = """
include:
  - includetest.include-two
    """

    include_one_sls = """
/tmp/from-test-file1.txt:
  file.managed:
    - contents: Hello from test-file.sls
    """

    include_two_sls = """
/tmp/from-test-file2.txt:
  file.managed:
    - contents: Hello from test-file.sls
    """

    init_sls = """
include:
  - includetest.only-include-one
  - includetest.only-include-two

/tmp/from-init.txt:
  file.managed:
    - contents: Hello from init.sls
    - require:
      - sls: includetest.only-include-one
      - sls: includetest.only-include-two
    """

    tf = salt_master.state_tree.base.temp_file

    with tf("includetest/init.sls", init_sls):
        with tf("includetest/include-one.sls", include_one_sls), tf(
            "includetest/include-two.sls", include_two_sls
        ):
            with tf("includetest/only-include-one.sls", only_include_one_sls), tf(
                "includetest/only-include-two.sls", only_include_two_sls
            ):
                ret = salt_call_cli.run("state.apply", "includetest")
                assert ret.returncode == 0

                ret = salt_call_cli.run("state.show_low_sls", "includetest")
                assert "__sls_included_from__" in ret.data[0]
                assert (
                    "includetest.only-include-one"
                    in ret.data[0]["__sls_included_from__"]
                )

                assert "__sls_included_from__" in ret.data[1]
                assert (
                    "includetest.only-include-two"
                    in ret.data[1]["__sls_included_from__"]
                )


@pytest.mark.slow_test
def test_issue_65080_saltenv(salt_master, salt_minion, salt_call_cli):
    """
    Test scenario when a state includes another state that only includes a third state
    """

    only_include_sls = """
include:
  - includetest.another-include
    """

    another_include_sls = """
/tmp/from-test-file.txt:
  file.managed:
    - contents: Hello from test-file.sls
    """

    init_sls = """
include:
  - prod: includetest.only-include

/tmp/from-init.txt:
  file.managed:
    - contents: Hello from init.sls
    - require:
      - sls: includetest.only-include
    """

    base_tf = salt_master.state_tree.base.temp_file
    prod_tf = salt_master.state_tree.prod.temp_file

    with base_tf("includetest/init.sls", init_sls):
        with prod_tf("includetest/only-include.sls", only_include_sls):
            with prod_tf("includetest/another-include.sls", another_include_sls):
                ret = salt_call_cli.run("state.apply", "includetest")
                assert ret.returncode == 0

                ret = salt_call_cli.run("state.show_low_sls", "includetest")
                assert "__sls_included_from__" in ret.data[0]
                assert (
                    "includetest.only-include" in ret.data[0]["__sls_included_from__"]
                )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.smtp as smtp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {smtp: {}}


def test_send_msg():
    """
    Test to send a message via SMTP
    """
    name = "This is a salt states module"

    comt = "Need to send message to admin@example.com: This is a salt states module"

    ret = {"name": name, "changes": {}, "result": None, "comment": comt}

    with patch.dict(smtp.__opts__, {"test": True}):
        assert ret == smtp.send_msg(
            name,
            "admin@example.com",
            "Message from Salt",
            "admin@example.com",
            "my-smtp-account",
        )

    comt = "Sent message to admin@example.com: This is a salt states module"

    with patch.dict(smtp.__opts__, {"test": False}):
        mock = MagicMock(return_value=True)
        with patch.dict(smtp.__salt__, {"smtp.send_msg": mock}):
            ret["comment"] = comt
            ret["result"] = True
            assert ret == smtp.send_msg(
                name,
                "admin@example.com",
                "Message from Salt",
                "admin@example.com",
                "my-smtp-account",
            )



"""
Compatibility functions for utils
"""

import copy
import importlib
import sys
import types

import salt.loader


def pack_dunder(name):
    """
    Compatibility helper function to make __utils__ available on demand.
    """
    # TODO: Deprecate starting with Beryllium

    mod = sys.modules[name]
    if not hasattr(mod, "__utils__"):
        setattr(
            mod, "__utils__", salt.loader.utils(mod.__opts__, pack_self="__utils__")
        )


def deepcopy_bound(name):
    """
    Compatibility helper function to allow copy.deepcopy copy bound methods
    which is broken on Python 2.6, due to the following bug:
    https://bugs.python.org/issue1515

    Warnings:
        - This method will mutate the global deepcopy dispatcher, which means that
        this function is NOT threadsafe!

        - Not Py3 compatible. The intended use case is deepcopy compat for Py2.6

    """

    def _deepcopy_method(x, memo):
        return type(x)(x.im_func, copy.deepcopy(x.im_self, memo), x.im_class)

    pre_dispatch = copy._deepcopy_dispatch
    try:
        copy._deepcopy_dispatch[types.MethodType] = _deepcopy_method
        ret = copy.deepcopy(name)
    finally:
        copy._deepcopy_dispatch = pre_dispatch
    return ret


def cmp(x, y):
    """
    Compatibility helper function to replace the ``cmp`` function from Python 2. The
    ``cmp`` function is no longer available in Python 3.

    cmp(x, y) -> integer

    Return negative if x<y, zero if x==y, positive if x>y.
    """
    if isinstance(x, dict) and isinstance(y, dict):
        return 0 if x == y else -1
    return (x > y) - (x < y)


def reload(mod):
    """
    Compatibility helper function to replace the ``reload`` builtin from Python 2.
    """
    try:
        return importlib.reload(mod)
    except AttributeError:
        return reload(mod)



template_engine_file_str = "no feature ospf"

config_result = [["no feature ospf"], [{}]]

initial_config = [
    "\n!Command: show running-config\n!Running configuration last done at: Wed Dec  4"
    " 15:33:48 2019\n!Time: Wed Dec  4 15:33:55 2019\n\nversion 9.2(4) Bios:version"
    " 07.66 \nhostname n9k-device\ninstall feature-set fex\nvdc n9k-device id 1\n "
    " allow feature-set fex\n  limit-resource vlan minimum 16 maximum 4094\n "
    " limit-resource vrf minimum 2 maximum 4096\n  limit-resource port-channel minimum"
    " 0 maximum 256\n  limit-resource u4route-mem minimum 248 maximum 248\n "
    " limit-resource u6route-mem minimum 96 maximum 96\n  limit-resource m4route-mem"
    " minimum 58 maximum 58\n  limit-resource m6route-mem minimum 8 maximum"
    " 8\nfeature-set fex\n\nfeature privilege\nfeature telnet\nfeature nxapi\nfeature"
    " bash-shell\ncfs eth distribute\nfeature ngmvpn\nfeature ospf\nfeature"
    " pim\nfeature lldp\n\nusername admin password 5"
    " $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role"
    " network-admin\nusername devops_user password 5 !  role network-operator\nusername"
    " devops_user passphrase  lifetime 99999 warntime 14 gracetime 3\nusername"
    " salt_test password 5 $5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC "
    " role network-operator\nusername salt_test passphrase  lifetime 99999 warntime 14"
    " gracetime 3\nip domain-lookup\nip name-server 1.1.1.1 2.2.2.2\nno system default"
    " switchport\nsystem default switchport shutdown\nno logging event link-status"
    " default\nno logging event link-status enable\nno logging event trunk-status"
    " enable\ntime-range ans-range\ntime-range my_range\ncopp profile"
    " strict\nsnmp-server user admin network-admin auth md5"
    " 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8"
    " localizedkey\nsnmp-server user devops_user network-operator \nsnmp-server user"
    " salt_test network-operator auth md5 0x3a16a4d0a569be282cc5fbc96138fac8 priv"
    " 0x3a16a4d0a569be282cc5fbc96138fac8 localizedkey\nrmon event 1 description"
    " FATAL(1) owner PMON@FATAL\nrmon event 2 description CRITICAL(2) owner"
    " PMON@CRITICAL\nrmon event 3 description ERROR(3) owner PMON@ERROR\nrmon event 4"
    " description WARNING(4) owner PMON@WARNING\nrmon event 5 description"
    " INFORMATION(5) owner PMON@INFO\nno snmp-server enable traps entity"
    " entity_mib_change\nno snmp-server enable traps entity"
    " entity_module_status_change\nno snmp-server enable traps entity"
    " entity_power_status_change\nno snmp-server enable traps entity"
    " entity_module_inserted\nno snmp-server enable traps entity"
    " entity_module_removed\nno snmp-server enable traps entity"
    " entity_unrecognised_module\nno snmp-server enable traps entity"
    " entity_fan_status_change\nno snmp-server enable traps entity"
    " entity_power_out_change\nno snmp-server enable traps link linkDown\nno"
    " snmp-server enable traps link linkUp\nno snmp-server enable traps link"
    " extended-linkDown\nno snmp-server enable traps link extended-linkUp\nno"
    " snmp-server enable traps link cieLinkDown\nno snmp-server enable traps link"
    " cieLinkUp\nno snmp-server enable traps link delayed-link-state-change\nno"
    " snmp-server enable traps rf redundancy_framework\nno snmp-server enable traps"
    " license notify-license-expiry\nno snmp-server enable traps license"
    " notify-no-license-for-feature\nno snmp-server enable traps license"
    " notify-licensefile-missing\nno snmp-server enable traps license"
    " notify-license-expiry-warning\nno snmp-server enable traps upgrade"
    " UpgradeOpNotifyOnCompletion\nno snmp-server enable traps upgrade"
    " UpgradeJobStatusNotify\nno snmp-server enable traps rmon risingAlarm\nno"
    " snmp-server enable traps rmon fallingAlarm\nno snmp-server enable traps rmon"
    " hcRisingAlarm\nno snmp-server enable traps rmon hcFallingAlarm\nno snmp-server"
    " enable traps entity entity_sensor\nno snmp-server enable traps entity"
    " cefcMIBEnableStatusNotification\nno snmp-server enable traps generic"
    " coldStart\nno snmp-server enable traps generic warmStart\nno snmp-server enable"
    " traps storm-control cpscEventRev1\nno snmp-server enable traps link"
    " cErrDisableInterfaceEventRev1\nno snmp-server enable traps link"
    " cmn-mac-move-notification\nntp server 10.81.254.202 use-vrf management\n\nip pim"
    " ssm range none\nvlan 1\n\nvrf context management\n  ip route 0.0.0.0/0"
    " 192.168.0.1\nvrf context myvrf\nvrf context testing\nnxapi http port 80\nno nxapi"
    " https\nnxapi ssl ciphers weak\nnxapi ssl protocols TLSv1 \n\n\ninterface"
    " port-channel19\n\ninterface port-channel42\n\ninterface"
    " port-channel55\n\ninterface port-channel66\n\ninterface Ethernet1/1\n\ninterface"
    " Ethernet1/2\n\ninterface Ethernet1/3\n  no shutdown\n\ninterface"
    " Ethernet1/4\n\ninterface Ethernet1/5\n\ninterface Ethernet1/6\n\ninterface"
    " Ethernet1/7\n\ninterface Ethernet1/8\n\ninterface Ethernet1/9\n\ninterface"
    " Ethernet1/10\n\ninterface Ethernet1/11\n\ninterface Ethernet1/12\n\ninterface"
    " Ethernet1/13\n\ninterface Ethernet1/14\n\ninterface Ethernet1/15\n\ninterface"
    " Ethernet1/16\n\ninterface Ethernet1/17\n\ninterface Ethernet1/18\n\ninterface"
    " Ethernet1/19\n\ninterface Ethernet1/20\n\ninterface Ethernet1/21\n\ninterface"
    " Ethernet1/22\n\ninterface Ethernet1/23\n\ninterface Ethernet1/24\n\ninterface"
    " Ethernet1/25\n\ninterface Ethernet1/26\n\ninterface Ethernet1/27\n\ninterface"
    " Ethernet1/28\n\ninterface Ethernet1/29\n\ninterface Ethernet1/30\n\ninterface"
    " Ethernet1/31\n\ninterface Ethernet1/32\n\ninterface Ethernet1/33\n\ninterface"
    " Ethernet1/34\n\ninterface Ethernet1/35\n\ninterface Ethernet1/36\n\ninterface"
    " Ethernet1/37\n\ninterface Ethernet1/38\n\ninterface Ethernet1/39\n\ninterface"
    " Ethernet1/40\n\ninterface Ethernet1/41\n\ninterface Ethernet1/42\n\ninterface"
    " Ethernet1/43\n\ninterface Ethernet1/44\n\ninterface Ethernet1/45\n\ninterface"
    " Ethernet1/46\n\ninterface Ethernet1/47\n\ninterface Ethernet1/48\n\ninterface"
    " Ethernet2/1\n\ninterface Ethernet2/2\n\ninterface Ethernet2/3\n\ninterface"
    " Ethernet2/4\n\ninterface Ethernet2/5\n\ninterface Ethernet2/6\n\ninterface"
    " Ethernet2/7\n\ninterface Ethernet2/8\n\ninterface Ethernet2/9\n\ninterface"
    " Ethernet2/10\n\ninterface Ethernet2/11\n\ninterface Ethernet2/12\n\ninterface"
    " mgmt0\n  vrf member management\n  ip address 192.168.0.192/24\n\ninterface"
    " loopback0\n\ninterface loopback1\ncli alias name wr copy running"
    " startup-config\nline console\nline vty\nboot nxos bootflash:/nxos.9.2.4.bin"
    " \n\nno logging logfile\nno logging monitor\nno logging module\nno logging"
    " console\n\n\n"
]

modified_config = [
    "\n!Command: show running-config\n!Running configuration last done at: Wed Dec  4"
    " 15:33:56 2019\n!Time: Wed Dec  4 15:33:56 2019\n\nversion 9.2(4) Bios:version"
    " 07.66 \nhostname n9k-device\ninstall feature-set fex\nvdc n9k-device id 1\n "
    " allow feature-set fex\n  limit-resource vlan minimum 16 maximum 4094\n "
    " limit-resource vrf minimum 2 maximum 4096\n  limit-resource port-channel minimum"
    " 0 maximum 256\n  limit-resource u4route-mem minimum 248 maximum 248\n "
    " limit-resource u6route-mem minimum 96 maximum 96\n  limit-resource m4route-mem"
    " minimum 58 maximum 58\n  limit-resource m6route-mem minimum 8 maximum"
    " 8\nfeature-set fex\n\nfeature privilege\nfeature telnet\nfeature nxapi\nfeature"
    " bash-shell\ncfs eth distribute\nfeature ngmvpn\nfeature pim\nfeature"
    " lldp\n\nusername admin password 5"
    " $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role"
    " network-admin\nusername devops_user password 5 !  role network-operator\nusername"
    " devops_user passphrase  lifetime 99999 warntime 14 gracetime 3\nusername"
    " salt_test password 5 $5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC "
    " role network-operator\nusername salt_test passphrase  lifetime 99999 warntime 14"
    " gracetime 3\nip domain-lookup\nip name-server 1.1.1.1 2.2.2.2\nno system default"
    " switchport\nsystem default switchport shutdown\nno logging event link-status"
    " default\nno logging event link-status enable\nno logging event trunk-status"
    " enable\ntime-range ans-range\ntime-range my_range\ncopp profile"
    " strict\nsnmp-server user admin network-admin auth md5"
    " 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8"
    " localizedkey\nsnmp-server user devops_user network-operator \nsnmp-server user"
    " salt_test network-operator auth md5 0x3a16a4d0a569be282cc5fbc96138fac8 priv"
    " 0x3a16a4d0a569be282cc5fbc96138fac8 localizedkey\nrmon event 1 description"
    " FATAL(1) owner PMON@FATAL\nrmon event 2 description CRITICAL(2) owner"
    " PMON@CRITICAL\nrmon event 3 description ERROR(3) owner PMON@ERROR\nrmon event 4"
    " description WARNING(4) owner PMON@WARNING\nrmon event 5 description"
    " INFORMATION(5) owner PMON@INFO\nno snmp-server enable traps entity"
    " entity_mib_change\nno snmp-server enable traps entity"
    " entity_module_status_change\nno snmp-server enable traps entity"
    " entity_power_status_change\nno snmp-server enable traps entity"
    " entity_module_inserted\nno snmp-server enable traps entity"
    " entity_module_removed\nno snmp-server enable traps entity"
    " entity_unrecognised_module\nno snmp-server enable traps entity"
    " entity_fan_status_change\nno snmp-server enable traps entity"
    " entity_power_out_change\nno snmp-server enable traps link linkDown\nno"
    " snmp-server enable traps link linkUp\nno snmp-server enable traps link"
    " extended-linkDown\nno snmp-server enable traps link extended-linkUp\nno"
    " snmp-server enable traps link cieLinkDown\nno snmp-server enable traps link"
    " cieLinkUp\nno snmp-server enable traps link delayed-link-state-change\nno"
    " snmp-server enable traps rf redundancy_framework\nno snmp-server enable traps"
    " license notify-license-expiry\nno snmp-server enable traps license"
    " notify-no-license-for-feature\nno snmp-server enable traps license"
    " notify-licensefile-missing\nno snmp-server enable traps license"
    " notify-license-expiry-warning\nno snmp-server enable traps upgrade"
    " UpgradeOpNotifyOnCompletion\nno snmp-server enable traps upgrade"
    " UpgradeJobStatusNotify\nno snmp-server enable traps rmon risingAlarm\nno"
    " snmp-server enable traps rmon fallingAlarm\nno snmp-server enable traps rmon"
    " hcRisingAlarm\nno snmp-server enable traps rmon hcFallingAlarm\nno snmp-server"
    " enable traps entity entity_sensor\nno snmp-server enable traps entity"
    " cefcMIBEnableStatusNotification\nno snmp-server enable traps generic"
    " coldStart\nno snmp-server enable traps generic warmStart\nno snmp-server enable"
    " traps storm-control cpscEventRev1\nno snmp-server enable traps link"
    " cErrDisableInterfaceEventRev1\nno snmp-server enable traps link"
    " cmn-mac-move-notification\nntp server 10.81.254.202 use-vrf management\n\nip pim"
    " ssm range none\nvlan 1\n\nvrf context management\n  ip route 0.0.0.0/0"
    " 192.168.0.1\nvrf context myvrf\nvrf context testing\nnxapi http port 80\nno nxapi"
    " https\nnxapi ssl ciphers weak\nnxapi ssl protocols TLSv1 \n\n\ninterface"
    " port-channel19\n\ninterface port-channel42\n\ninterface"
    " port-channel55\n\ninterface port-channel66\n\ninterface Ethernet1/1\n\ninterface"
    " Ethernet1/2\n\ninterface Ethernet1/3\n  no shutdown\n\ninterface"
    " Ethernet1/4\n\ninterface Ethernet1/5\n\ninterface Ethernet1/6\n\ninterface"
    " Ethernet1/7\n\ninterface Ethernet1/8\n\ninterface Ethernet1/9\n\ninterface"
    " Ethernet1/10\n\ninterface Ethernet1/11\n\ninterface Ethernet1/12\n\ninterface"
    " Ethernet1/13\n\ninterface Ethernet1/14\n\ninterface Ethernet1/15\n\ninterface"
    " Ethernet1/16\n\ninterface Ethernet1/17\n\ninterface Ethernet1/18\n\ninterface"
    " Ethernet1/19\n\ninterface Ethernet1/20\n\ninterface Ethernet1/21\n\ninterface"
    " Ethernet1/22\n\ninterface Ethernet1/23\n\ninterface Ethernet1/24\n\ninterface"
    " Ethernet1/25\n\ninterface Ethernet1/26\n\ninterface Ethernet1/27\n\ninterface"
    " Ethernet1/28\n\ninterface Ethernet1/29\n\ninterface Ethernet1/30\n\ninterface"
    " Ethernet1/31\n\ninterface Ethernet1/32\n\ninterface Ethernet1/33\n\ninterface"
    " Ethernet1/34\n\ninterface Ethernet1/35\n\ninterface Ethernet1/36\n\ninterface"
    " Ethernet1/37\n\ninterface Ethernet1/38\n\ninterface Ethernet1/39\n\ninterface"
    " Ethernet1/40\n\ninterface Ethernet1/41\n\ninterface Ethernet1/42\n\ninterface"
    " Ethernet1/43\n\ninterface Ethernet1/44\n\ninterface Ethernet1/45\n\ninterface"
    " Ethernet1/46\n\ninterface Ethernet1/47\n\ninterface Ethernet1/48\n\ninterface"
    " Ethernet2/1\n\ninterface Ethernet2/2\n\ninterface Ethernet2/3\n\ninterface"
    " Ethernet2/4\n\ninterface Ethernet2/5\n\ninterface Ethernet2/6\n\ninterface"
    " Ethernet2/7\n\ninterface Ethernet2/8\n\ninterface Ethernet2/9\n\ninterface"
    " Ethernet2/10\n\ninterface Ethernet2/11\n\ninterface Ethernet2/12\n\ninterface"
    " mgmt0\n  vrf member management\n  ip address 192.168.0.192/24\n\ninterface"
    " loopback0\n\ninterface loopback1\ncli alias name wr copy running"
    " startup-config\nline console\nline vty\nboot nxos bootflash:/nxos.9.2.4.bin"
    " \n\nno logging logfile\nno logging monitor\nno logging module\nno logging"
    " console\n\n\n"
]

config_input_file = """
feature bgp
!
router bgp 55
  address-family ipv4 unicast
    no client-to-client reflection
    additional-paths send
"""

template_engine_file_str_file = config_input_file

config_result_file = [
    [
        "feature bgp",
        "!",
        "router bgp 55",
        "  address-family ipv4 unicast",
        "    no client-to-client reflection",
        "    additional-paths send",
    ],
    [{}, {}, {}, {}, {}, {}],
]

initial_config_file = [
    "\n!Command: show running-config\n!Running configuration last done at: Wed Dec  4"
    " 15:33:56 2019\n!Time: Wed Dec  4 15:55:51 2019\n\nversion 9.2(4) Bios:version"
    " 07.66 \nhostname dt-n9k5-1\ninstall feature-set fex\nvdc dt-n9k5-1 id 1\n  allow"
    " feature-set fex\n  limit-resource vlan minimum 16 maximum 4094\n  limit-resource"
    " vrf minimum 2 maximum 4096\n  limit-resource port-channel minimum 0 maximum 256\n"
    "  limit-resource u4route-mem minimum 248 maximum 248\n  limit-resource u6route-mem"
    " minimum 96 maximum 96\n  limit-resource m4route-mem minimum 58 maximum 58\n "
    " limit-resource m6route-mem minimum 8 maximum 8\nfeature-set fex\n\nfeature"
    " privilege\nfeature telnet\nfeature nxapi\nfeature bash-shell\ncfs eth"
    " distribute\nfeature ngmvpn\nfeature pim\nfeature lldp\n\nusername admin password"
    " 5 $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role"
    " network-admin\nusername devops_user password 5 !  role network-operator\nusername"
    " devops_user passphrase  lifetime 99999 warntime 14 gracetime 3\nusername"
    " salt_test password 5 $5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC "
    " role network-operator\nusername salt_test passphrase  lifetime 99999 warntime 14"
    " gracetime 3\nip domain-lookup\nip name-server 1.1.1.1 2.2.2.2\nno system default"
    " switchport\nsystem default switchport shutdown\nno logging event link-status"
    " default\nno logging event link-status enable\nno logging event trunk-status"
    " enable\ntime-range ans-range\ntime-range my_range\ncopp profile"
    " strict\nsnmp-server user admin network-admin auth md5"
    " 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8"
    " localizedkey\nsnmp-server user devops_user network-operator \nsnmp-server user"
    " salt_test network-operator auth md5 0x3a16a4d0a569be282cc5fbc96138fac8 priv"
    " 0x3a16a4d0a569be282cc5fbc96138fac8 localizedkey\nrmon event 1 description"
    " FATAL(1) owner PMON@FATAL\nrmon event 2 description CRITICAL(2) owner"
    " PMON@CRITICAL\nrmon event 3 description ERROR(3) owner PMON@ERROR\nrmon event 4"
    " description WARNING(4) owner PMON@WARNING\nrmon event 5 description"
    " INFORMATION(5) owner PMON@INFO\nno snmp-server enable traps entity"
    " entity_mib_change\nno snmp-server enable traps entity"
    " entity_module_status_change\nno snmp-server enable traps entity"
    " entity_power_status_change\nno snmp-server enable traps entity"
    " entity_module_inserted\nno snmp-server enable traps entity"
    " entity_module_removed\nno snmp-server enable traps entity"
    " entity_unrecognised_module\nno snmp-server enable traps entity"
    " entity_fan_status_change\nno snmp-server enable traps entity"
    " entity_power_out_change\nno snmp-server enable traps link linkDown\nno"
    " snmp-server enable traps link linkUp\nno snmp-server enable traps link"
    " extended-linkDown\nno snmp-server enable traps link extended-linkUp\nno"
    " snmp-server enable traps link cieLinkDown\nno snmp-server enable traps link"
    " cieLinkUp\nno snmp-server enable traps link delayed-link-state-change\nno"
    " snmp-server enable traps rf redundancy_framework\nno snmp-server enable traps"
    " license notify-license-expiry\nno snmp-server enable traps license"
    " notify-no-license-for-feature\nno snmp-server enable traps license"
    " notify-licensefile-missing\nno snmp-server enable traps license"
    " notify-license-expiry-warning\nno snmp-server enable traps upgrade"
    " UpgradeOpNotifyOnCompletion\nno snmp-server enable traps upgrade"
    " UpgradeJobStatusNotify\nno snmp-server enable traps rmon risingAlarm\nno"
    " snmp-server enable traps rmon fallingAlarm\nno snmp-server enable traps rmon"
    " hcRisingAlarm\nno snmp-server enable traps rmon hcFallingAlarm\nno snmp-server"
    " enable traps entity entity_sensor\nno snmp-server enable traps entity"
    " cefcMIBEnableStatusNotification\nno snmp-server enable traps generic"
    " coldStart\nno snmp-server enable traps generic warmStart\nno snmp-server enable"
    " traps storm-control cpscEventRev1\nno snmp-server enable traps link"
    " cErrDisableInterfaceEventRev1\nno snmp-server enable traps link"
    " cmn-mac-move-notification\nntp server 10.81.254.202 use-vrf management\n\nip pim"
    " ssm range none\nvlan 1\n\nvrf context management\n  ip route 0.0.0.0/0"
    " 10.122.197.1\nvrf context myvrf\nvrf context testing\nnxapi http port 80\nno"
    " nxapi https\nnxapi ssl ciphers weak\nnxapi ssl protocols TLSv1 \n\n\ninterface"
    " port-channel19\n\ninterface port-channel42\n\ninterface"
    " port-channel55\n\ninterface port-channel66\n\ninterface Ethernet1/1\n\ninterface"
    " Ethernet1/2\n\ninterface Ethernet1/3\n  no shutdown\n\ninterface"
    " Ethernet1/4\n\ninterface Ethernet1/5\n\ninterface Ethernet1/6\n\ninterface"
    " Ethernet1/7\n\ninterface Ethernet1/8\n\ninterface Ethernet1/9\n\ninterface"
    " Ethernet1/10\n\ninterface Ethernet1/11\n\ninterface Ethernet1/12\n\ninterface"
    " Ethernet1/13\n\ninterface Ethernet1/14\n\ninterface Ethernet1/15\n\ninterface"
    " Ethernet1/16\n\ninterface Ethernet1/17\n\ninterface Ethernet1/18\n\ninterface"
    " Ethernet1/19\n\ninterface Ethernet1/20\n\ninterface Ethernet1/21\n\ninterface"
    " Ethernet1/22\n\ninterface Ethernet1/23\n\ninterface Ethernet1/24\n\ninterface"
    " Ethernet1/25\n\ninterface Ethernet1/26\n\ninterface Ethernet1/27\n\ninterface"
    " Ethernet1/28\n\ninterface Ethernet1/29\n\ninterface Ethernet1/30\n\ninterface"
    " Ethernet1/31\n\ninterface Ethernet1/32\n\ninterface Ethernet1/33\n\ninterface"
    " Ethernet1/34\n\ninterface Ethernet1/35\n\ninterface Ethernet1/36\n\ninterface"
    " Ethernet1/37\n\ninterface Ethernet1/38\n\ninterface Ethernet1/39\n\ninterface"
    " Ethernet1/40\n\ninterface Ethernet1/41\n\ninterface Ethernet1/42\n\ninterface"
    " Ethernet1/43\n\ninterface Ethernet1/44\n\ninterface Ethernet1/45\n\ninterface"
    " Ethernet1/46\n\ninterface Ethernet1/47\n\ninterface Ethernet1/48\n\ninterface"
    " Ethernet2/1\n\ninterface Ethernet2/2\n\ninterface Ethernet2/3\n\ninterface"
    " Ethernet2/4\n\ninterface Ethernet2/5\n\ninterface Ethernet2/6\n\ninterface"
    " Ethernet2/7\n\ninterface Ethernet2/8\n\ninterface Ethernet2/9\n\ninterface"
    " Ethernet2/10\n\ninterface Ethernet2/11\n\ninterface Ethernet2/12\n\ninterface"
    " mgmt0\n  vrf member management\n  ip address 10.122.197.192/24\n\ninterface"
    " loopback0\n\ninterface loopback1\ncli alias name wr copy running"
    " startup-config\nline console\nline vty\nboot nxos bootflash:/nxos.9.2.4.bin"
    " \n\nno logging logfile\nno logging monitor\nno logging module\nno logging"
    " console\n\n\n"
]

modified_config_file = [
    "\n!Command: show running-config\n!Running configuration last done at: Wed Dec  4"
    " 15:55:55 2019\n!Time: Wed Dec  4 15:55:55 2019\n\nversion 9.2(4) Bios:version"
    " 07.66 \nhostname dt-n9k5-1\ninstall feature-set fex\nvdc dt-n9k5-1 id 1\n  allow"
    " feature-set fex\n  limit-resource vlan minimum 16 maximum 4094\n  limit-resource"
    " vrf minimum 2 maximum 4096\n  limit-resource port-channel minimum 0 maximum 256\n"
    "  limit-resource u4route-mem minimum 248 maximum 248\n  limit-resource u6route-mem"
    " minimum 96 maximum 96\n  limit-resource m4route-mem minimum 58 maximum 58\n "
    " limit-resource m6route-mem minimum 8 maximum 8\nfeature-set fex\n\nfeature"
    " privilege\nfeature telnet\nfeature nxapi\nfeature bash-shell\ncfs eth"
    " distribute\nfeature ngmvpn\nfeature bgp\nfeature pim\nfeature lldp\n\nusername"
    " admin password 5 $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role"
    " network-admin\nusername devops_user password 5 !  role network-operator\nusername"
    " devops_user passphrase  lifetime 99999 warntime 14 gracetime 3\nusername"
    " salt_test password 5 $5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC "
    " role network-operator\nusername salt_test passphrase  lifetime 99999 warntime 14"
    " gracetime 3\nip domain-lookup\nip name-server 1.1.1.1 2.2.2.2\nno system default"
    " switchport\nsystem default switchport shutdown\nno logging event link-status"
    " default\nno logging event link-status enable\nno logging event trunk-status"
    " enable\ntime-range ans-range\ntime-range my_range\ncopp profile"
    " strict\nsnmp-server user admin network-admin auth md5"
    " 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8"
    " localizedkey\nsnmp-server user devops_user network-operator \nsnmp-server user"
    " salt_test network-operator auth md5 0x3a16a4d0a569be282cc5fbc96138fac8 priv"
    " 0x3a16a4d0a569be282cc5fbc96138fac8 localizedkey\nrmon event 1 description"
    " FATAL(1) owner PMON@FATAL\nrmon event 2 description CRITICAL(2) owner"
    " PMON@CRITICAL\nrmon event 3 description ERROR(3) owner PMON@ERROR\nrmon event 4"
    " description WARNING(4) owner PMON@WARNING\nrmon event 5 description"
    " INFORMATION(5) owner PMON@INFO\nno snmp-server enable traps entity"
    " entity_mib_change\nno snmp-server enable traps entity"
    " entity_module_status_change\nno snmp-server enable traps entity"
    " entity_power_status_change\nno snmp-server enable traps entity"
    " entity_module_inserted\nno snmp-server enable traps entity"
    " entity_module_removed\nno snmp-server enable traps entity"
    " entity_unrecognised_module\nno snmp-server enable traps entity"
    " entity_fan_status_change\nno snmp-server enable traps entity"
    " entity_power_out_change\nno snmp-server enable traps link linkDown\nno"
    " snmp-server enable traps link linkUp\nno snmp-server enable traps link"
    " extended-linkDown\nno snmp-server enable traps link extended-linkUp\nno"
    " snmp-server enable traps link cieLinkDown\nno snmp-server enable traps link"
    " cieLinkUp\nno snmp-server enable traps link delayed-link-state-change\nno"
    " snmp-server enable traps rf redundancy_framework\nno snmp-server enable traps"
    " license notify-license-expiry\nno snmp-server enable traps license"
    " notify-no-license-for-feature\nno snmp-server enable traps license"
    " notify-licensefile-missing\nno snmp-server enable traps license"
    " notify-license-expiry-warning\nno snmp-server enable traps upgrade"
    " UpgradeOpNotifyOnCompletion\nno snmp-server enable traps upgrade"
    " UpgradeJobStatusNotify\nno snmp-server enable traps rmon risingAlarm\nno"
    " snmp-server enable traps rmon fallingAlarm\nno snmp-server enable traps rmon"
    " hcRisingAlarm\nno snmp-server enable traps rmon hcFallingAlarm\nno snmp-server"
    " enable traps entity entity_sensor\nno snmp-server enable traps entity"
    " cefcMIBEnableStatusNotification\nno snmp-server enable traps generic"
    " coldStart\nno snmp-server enable traps generic warmStart\nno snmp-server enable"
    " traps storm-control cpscEventRev1\nno snmp-server enable traps link"
    " cErrDisableInterfaceEventRev1\nno snmp-server enable traps link"
    " cmn-mac-move-notification\nntp server 10.81.254.202 use-vrf management\n\nip pim"
    " ssm range none\nvlan 1\n\nvrf context management\n  ip route 0.0.0.0/0"
    " 10.122.197.1\nvrf context myvrf\nvrf context testing\nnxapi http port 80\nno"
    " nxapi https\nnxapi ssl ciphers weak\nnxapi ssl protocols TLSv1 \n\n\ninterface"
    " port-channel19\n\ninterface port-channel42\n\ninterface"
    " port-channel55\n\ninterface port-channel66\n\ninterface Ethernet1/1\n\ninterface"
    " Ethernet1/2\n\ninterface Ethernet1/3\n  no shutdown\n\ninterface"
    " Ethernet1/4\n\ninterface Ethernet1/5\n\ninterface Ethernet1/6\n\ninterface"
    " Ethernet1/7\n\ninterface Ethernet1/8\n\ninterface Ethernet1/9\n\ninterface"
    " Ethernet1/10\n\ninterface Ethernet1/11\n\ninterface Ethernet1/12\n\ninterface"
    " Ethernet1/13\n\ninterface Ethernet1/14\n\ninterface Ethernet1/15\n\ninterface"
    " Ethernet1/16\n\ninterface Ethernet1/17\n\ninterface Ethernet1/18\n\ninterface"
    " Ethernet1/19\n\ninterface Ethernet1/20\n\ninterface Ethernet1/21\n\ninterface"
    " Ethernet1/22\n\ninterface Ethernet1/23\n\ninterface Ethernet1/24\n\ninterface"
    " Ethernet1/25\n\ninterface Ethernet1/26\n\ninterface Ethernet1/27\n\ninterface"
    " Ethernet1/28\n\ninterface Ethernet1/29\n\ninterface Ethernet1/30\n\ninterface"
    " Ethernet1/31\n\ninterface Ethernet1/32\n\ninterface Ethernet1/33\n\ninterface"
    " Ethernet1/34\n\ninterface Ethernet1/35\n\ninterface Ethernet1/36\n\ninterface"
    " Ethernet1/37\n\ninterface Ethernet1/38\n\ninterface Ethernet1/39\n\ninterface"
    " Ethernet1/40\n\ninterface Ethernet1/41\n\ninterface Ethernet1/42\n\ninterface"
    " Ethernet1/43\n\ninterface Ethernet1/44\n\ninterface Ethernet1/45\n\ninterface"
    " Ethernet1/46\n\ninterface Ethernet1/47\n\ninterface Ethernet1/48\n\ninterface"
    " Ethernet2/1\n\ninterface Ethernet2/2\n\ninterface Ethernet2/3\n\ninterface"
    " Ethernet2/4\n\ninterface Ethernet2/5\n\ninterface Ethernet2/6\n\ninterface"
    " Ethernet2/7\n\ninterface Ethernet2/8\n\ninterface Ethernet2/9\n\ninterface"
    " Ethernet2/10\n\ninterface Ethernet2/11\n\ninterface Ethernet2/12\n\ninterface"
    " mgmt0\n  vrf member management\n  ip address 10.122.197.192/24\n\ninterface"
    " loopback0\n\ninterface loopback1\ncli alias name wr copy running"
    " startup-config\nline console\nline vty\nboot nxos bootflash:/nxos.9.2.4.bin"
    " \nrouter bgp 55\n  address-family ipv4 unicast\n    no client-to-client"
    " reflection\n    additional-paths send\n\nno logging logfile\nno logging"
    " monitor\nno logging module\nno logging console\n\n\n"
]

delete_config = """
COMMAND_LIST: no feature bgp

---
+++
@@ -19,7 +19,6 @@
 feature bash-shell
 cfs eth distribute
 feature ngmvpn
-feature bgp
 feature pim
 feature lldp

@@ -234,10 +233,6 @@
 line console
 line vty
 boot nxos bootflash:/nxos.9.2.4.bin
-router bgp 55
-  address-family ipv4 unicast
-    no client-to-client reflection
-    additional-paths send

 no logging logfile
 no logging monitor
 """

remove_user = """
COMMAND_LIST: no username salt_test

---
+++
@@ -25,8 +25,6 @@
 username admin password 5 $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role network-admin
 username devops_user password 5 !  role network-operator
 username devops_user passphrase  lifetime 99999 warntime 14 gracetime 3
-username salt_test password 5 $5$k8DulS4J$GGvG0RPTZdjklgJ4o6X25/AGpEHw1p1Dz8lI0.J77y3  role network-operator
-username salt_test passphrase  lifetime 99999 warntime 14 gracetime 3
 ip domain-lookup
 ip name-server 1.1.1.1 2.2.2.2
 no system default switchport
@@ -39,7 +37,6 @@
 copp profile strict
 snmp-server user admin network-admin auth md5 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8 localizedkey
 snmp-server user devops_user network-operator
-snmp-server user salt_test network-operator auth md5 0x6a4237693d3e515b67264a7872476132 priv 0x6a4237693d3e515b67264a7872476132 localizedkey
 rmon event 1 description FATAL(1) owner PMON@FATAL
 rmon event 2 description CRITICAL(2) owner PMON@CRITICAL
 rmon event 3 description ERROR(3) owner PMON@ERROR
 """

save_running_config = """
COMMAND_LIST: copy running-config startup-config

[#                                       ]   1%
[#                                       ]   2%
[##                                      ]   3%
[##                                      ]   4%
[###                                     ]   5%
[###                                     ]   6%
[###                                     ]   7%
[####                                    ]   8%
[####                                    ]   9%
[#####                                   ]  10%
[#####                                   ]  11%
[#####                                   ]  12%
[######                                  ]  13%
[######                                  ]  14%
[#######                                 ]  15%
[#######                                 ]  16%
[#######                                 ]  17%
[########                                ]  18%
[########                                ]  19%
[#########                               ]  20%
[#########                               ]  21%
[#########                               ]  22%
[##########                              ]  23%
[##########                              ]  24%
[###########                             ]  25%
[###########                             ]  26%
[###########                             ]  27%
[############                            ]  28%
[############                            ]  29%
[#############                           ]  30%
[#############                           ]  31%
[#############                           ]  32%
[##############                          ]  33%
[##############                          ]  34%
[###############                         ]  35%
[###############                         ]  36%
[###############                         ]  37%
[################                        ]  38%
[################                        ]  39%
[#################                       ]  40%
[#################                       ]  41%
[#################                       ]  42%
[##################                      ]  43%
[##################                      ]  44%
[###################                     ]  45%
[###################                     ]  46%
[###################                     ]  47%
[####################                    ]  48%
[#####################                   ]  50%
[#####################                   ]  51%
[#####################                   ]  52%
[######################                  ]  53%
[######################                  ]  54%
[#######################                 ]  55%
[#######################                 ]  56%
[#######################                 ]  57%
[########################                ]  58%
[########################                ]  59%
[#########################               ]  60%
[#########################               ]  61%
[#########################               ]  62%
[##########################              ]  63%
[##########################              ]  64%
[###########################             ]  65%
[###########################             ]  66%
[###########################             ]  67%
[############################            ]  68%
[############################            ]  69%
[#############################           ]  70%
[#############################           ]  71%
[#############################           ]  72%
[##############################          ]  73%
[##############################          ]  74%
[###############################         ]  75%
[###############################         ]  76%
[###############################         ]  77%
[################################        ]  78%
[################################        ]  79%
[#################################       ]  80%
[#################################       ]  81%
[#################################       ]  82%
[##################################      ]  83%
[##################################      ]  84%
[###################################     ]  85%
[###################################     ]  86%
[###################################     ]  87%
[####################################    ]  88%
[####################################    ]  89%
[#####################################   ]  90%
[#####################################   ]  91%
[#####################################   ]  92%
[######################################  ]  93%
[######################################  ]  94%
[####################################### ]  95%
[####################################### ]  96%
[####################################### ]  97%
[########################################]  98%
[########################################] 100%
Copy complete, now saving to disk (please wait)...
Copy complete.
"""

set_role = """
COMMAND_LIST: username salt_test role vdc-admin
warning: password for user:salt_test not set. S/he may not be able to login

---
+++
@@ -26,6 +26,8 @@
 username admin password 5 $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role network-admin
 username devops_user password 5 !  role network-operator
 username devops_user passphrase  lifetime 99999 warntime 14 gracetime 3
+username salt_test password 5 !  role vdc-admin
+username salt_test passphrase  lifetime 99999 warntime 14 gracetime 3
 ip domain-lookup
 ip name-server 1.1.1.1 2.2.2.2
 no system default switchport
@@ -38,6 +40,7 @@
 copp profile strict
 snmp-server user admin network-admin auth md5 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8 localizedkey
 snmp-server user devops_user network-operator
+snmp-server user salt_test vdc-admin
 rmon event 1 description FATAL(1) owner PMON@FATAL
 rmon event 2 description CRITICAL(2) owner PMON@CRITICAL
 rmon event 3 description ERROR(3) owner PMON@ERROR
"""

unset_role = """
COMMAND_LIST: no username salt_test role vdc-admin

---
+++
@@ -26,8 +26,7 @@
 username admin password 5 $5$mOJsBDrL$lhsuyA8VT/fkWM/5XaZcgHlJIYaaS/J5fh8qdDr6fU3  role network-admin
 username ansible password 5 !  role network-operator
 username ansible passphrase  lifetime 99999 warntime 14 gracetime 3
-username salt_test password 5 !  role vdc-admin
-username salt_test role network-operator
+username salt_test password 5 !  role network-operator
 username salt_test passphrase  lifetime 99999 warntime 14 gracetime 3
 ip domain-lookup
 ip name-server 1.1.1.1 2.2.2.2
@@ -41,8 +40,7 @@
 copp profile strict
 snmp-server user admin network-admin auth md5 0xbb62c3df89880933539503f7675e8aa8 priv 0xbb62c3df89880933539503f7675e8aa8 localizedkey
 snmp-server user ansible network-operator
-snmp-server user salt_test vdc-admin
-snmp-server user salt_test network-operator
+snmp-server user salt_test network-operator
 rmon event 1 description FATAL(1) owner PMON@FATAL
 rmon event 2 description CRITICAL(2) owner PMON@CRITICAL
 rmon event 3 description ERROR(3) owner PMON@ERROR
"""



"""
Return salt data via Telegram.

The following fields can be set in the minion conf file::

    telegram.chat_id (required)
    telegram.token (required)

Telegram settings may also be configured as:

.. code-block:: yaml

    telegram:
      chat_id: 000000000
      token: 000000000:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

To use the Telegram return, append '--return telegram' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return telegram

"""

import logging

import salt.returners

log = logging.getLogger(__name__)

__virtualname__ = "telegram"


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    return __virtualname__


def _get_options(ret=None):
    """
    Get the Telegram options from salt.

    :param ret:     The data to be sent.
    :return:        Dictionary containing the data and options needed to send
                    them to telegram.
    """
    attrs = {"chat_id": "chat_id", "token": "token"}

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )
    log.debug("Options: %s", _options)
    return _options


def returner(ret):
    """
    Send a Telegram message with the data.

    :param ret:     The data to be sent.
    :return:        Boolean if message was sent successfully.
    """
    _options = _get_options(ret)

    chat_id = _options.get("chat_id")
    token = _options.get("token")

    if not chat_id:
        log.error("telegram.chat_id not defined in salt config")
    if not token:
        log.error("telegram.token not defined in salt config")

    returns = ret.get("return")

    message = "id: {}\r\nfunction: {}\r\nfunction args: {}\r\njid: {}\r\nreturn: {}\r\n".format(
        ret.get("id"), ret.get("fun"), ret.get("fun_args"), ret.get("jid"), returns
    )

    return __salt__["telegram.post_message"](message, chat_id=chat_id, token=token)



"""
Tests for the win_pkg module
"""

import logging

import pytest

import salt.loader.dunder
import salt.modules.config as config
import salt.modules.cp as cp
import salt.modules.pkg_resource as pkg_resource
import salt.modules.win_pkg as win_pkg
import salt.utils.data
import salt.utils.platform
import salt.utils.win_reg as win_reg
from salt.exceptions import MinionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    pkg_info = {
        "latest": {
            "full_name": "Nullsoft Install System",
            "installer": "http://download.sourceforge.net/project/nsis/nsis-setup.exe",
            "install_flags": "/S",
            "uninstaller": "%PROGRAMFILES(x86)%\\NSIS\\uninst-nsis.exe",
            "uninstall_flags": "/S",
            "msiexec": False,
            "reboot": False,
        },
        "3.03": {
            "full_name": "Nullsoft Install System",
            "installer": "http://download.sourceforge.net/project/nsis/NSIS%203/3.03/nsis-3.03-setup.exe",
            "install_flags": "/S",
            "uninstaller": "%PROGRAMFILES(x86)%\\NSIS\\uninst-nsis.exe",
            "uninstall_flags": "/S",
            "msiexec": False,
            "reboot": False,
        },
        "3.02": {
            "full_name": "Nullsoft Install System",
            "installer": "http://download.sourceforge.net/project/nsis/NSIS%203/3.02/nsis-3.02-setup.exe",
            "install_flags": "/S",
            "uninstaller": "%PROGRAMFILES(x86)%\\NSIS\\uninst-nsis.exe",
            "uninstall_flags": "/S",
            "msiexec": False,
            "reboot": False,
        },
    }

    opts = minion_opts
    opts["master_uri"] = "localhost"
    return {
        cp: {"__opts__": salt.loader.dunder.__opts__.with_default(opts)},
        win_pkg: {
            "_get_latest_package_version": MagicMock(return_value="3.03"),
            "_get_package_info": MagicMock(return_value=pkg_info),
            "__salt__": {
                "config.valid_fileproto": config.valid_fileproto,
                "cp.hash_file": cp.hash_file,
                "pkg_resource.add_pkg": pkg_resource.add_pkg,
                "pkg_resource.parse_targets": pkg_resource.parse_targets,
                "pkg_resource.sort_pkglist": pkg_resource.sort_pkglist,
                "pkg_resource.stringify": pkg_resource.stringify,
            },
            "__utils__": {
                "reg.key_exists": win_reg.key_exists,
                "reg.list_keys": win_reg.list_keys,
                "reg.read_value": win_reg.read_value,
                "reg.value_exists": win_reg.value_exists,
            },
        },
        pkg_resource: {"__grains__": {"os": "Windows"}},
    }


def test_pkg__get_reg_software():
    result = win_pkg._get_reg_software()
    assert isinstance(result, dict)
    found_python = False
    search = "Python 3"
    for key in result:
        if search in key:
            found_python = True
    assert found_python


def test_pkg__get_reg_software_noremove():
    search = "test_pkg_noremove"
    key = f"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{search}"
    win_reg.set_value(hive="HKLM", key=key, vname="DisplayName", vdata=search)
    win_reg.set_value(hive="HKLM", key=key, vname="DisplayVersion", vdata="1.0.0")
    win_reg.set_value(
        hive="HKLM", key=key, vname="NoRemove", vtype="REG_DWORD", vdata="1"
    )
    try:
        result = win_pkg._get_reg_software()
        assert isinstance(result, dict)
        found = False
        search = "test_pkg"
        for item in result:
            if search in item:
                found = True
        assert found is True
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=key)
        assert not win_reg.key_exists(hive="HKLM", key=key)


def test_pkg__get_reg_software_noremove_not_present():
    search = "test_pkg_noremove_not_present"
    key = f"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{search}"
    win_reg.set_value(hive="HKLM", key=key, vname="DisplayName", vdata=search)
    win_reg.set_value(hive="HKLM", key=key, vname="DisplayVersion", vdata="1.0.0")
    try:
        result = win_pkg._get_reg_software()
        assert isinstance(result, dict)
        found = False
        for item in result:
            if search in item:
                found = True
        assert found is False
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=key)
        assert not win_reg.key_exists(hive="HKLM", key=key)


def test_pkg_install_not_found():
    """
    Test pkg.install when the Version is NOT FOUND in the Software
    Definition
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run with stringify
    se_list_pkgs = {"nsis": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ):
        expected = {"nsis": {"not found": "3.01"}}
        result = win_pkg.install(name="nsis", version="3.01")
        assert expected == result


def test_pkg_install_rollback():
    """
    test pkg.install rolling back to a previous version
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = [{"nsis": ["3.03"]}, {"nsis": "3.02"}]
    with patch.object(win_pkg, "list_pkgs", side_effect=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__, {"cp.is_cached": MagicMock(return_value=False)}
    ), patch.dict(
        win_pkg.__salt__,
        {"cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe")},
    ), patch.dict(
        win_pkg.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        expected = {"nsis": {"new": "3.02", "old": "3.03"}}
        result = win_pkg.install(name="nsis", version="3.02")
        assert expected == result


def test_pkg_install_existing():
    """
    test pkg.install when the package is already installed
    no version passed
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = {"nsis": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cmd.run_all": MagicMock(return_value={"retcode": 0}),
            "cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe"),
            "cp.is_cached": MagicMock(return_value=True),
        },
    ):
        expected = {}
        result = win_pkg.install(name="nsis")
        assert expected == result


def test_pkg_install_latest():
    """
    test pkg.install when the package is already installed
    no version passed
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = [{"nsis": ["3.03"]}, {"nsis": "3.04"}]
    mock_cache_file = MagicMock(return_value="C:\\fake\\path.exe")
    with patch.object(win_pkg, "list_pkgs", side_effect=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cmd.run_all": MagicMock(return_value={"retcode": 0}),
            "cp.cache_file": mock_cache_file,
            "cp.is_cached": MagicMock(return_value=False),
        },
    ):
        expected = {"nsis": {"new": "3.04", "old": "3.03"}}
        result = win_pkg.install(name="nsis", version="latest")
        assert expected == result
        mock_cache_file.assert_called_once_with(
            "http://download.sourceforge.net/project/nsis/nsis-setup.exe",
            saltenv="base",
            source_hash=None,
            verify_ssl=True,
            use_etag=True,
        )


def test_pkg_install_latest_is_cached():
    """
    test pkg.install when the package is already installed
    no version passed
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = [{"nsis": ["3.03"]}, {"nsis": "3.04"}]
    mock_cache_file = MagicMock(return_value="C:\\fake\\path.exe")
    with patch.object(win_pkg, "list_pkgs", side_effect=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cmd.run_all": MagicMock(return_value={"retcode": 0}),
            "cp.cache_file": mock_cache_file,
            "cp.is_cached": MagicMock(return_value=True),
        },
    ):
        expected = {"nsis": {"new": "3.04", "old": "3.03"}}
        result = win_pkg.install(name="nsis", version="latest")
        assert expected == result
        mock_cache_file.assert_called_once_with(
            "http://download.sourceforge.net/project/nsis/nsis-setup.exe",
            saltenv="base",
            source_hash=None,
            verify_ssl=True,
            use_etag=True,
        )


def test_pkg_install_existing_with_version():
    """
    test pkg.install when the package is already installed
    A version is passed
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = {"nsis": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cmd.run_all": MagicMock(return_value={"retcode": 0}),
            "cp.cache_file": MagicMock(return_value="C:\\fake\\path.exe"),
            "cp.is_cached": MagicMock(return_value=False),
        },
    ):
        expected = {}
        result = win_pkg.install(name="nsis", version="3.03")
        assert expected == result


def test_pkg_install_name():
    """
    test pkg.install name extra_install_flags
    """

    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": MagicMock(
                return_value=[{"firebox": "3.03"}, None]
            ),
            "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
            "cmd.run_all": mock_cmd_run_all,
        },
    ):
        win_pkg.install(
            name="firebox",
            version="3.03",
            extra_install_flags="-e True -test_flag True",
        )
        assert "-e True -test_flag True" in str(mock_cmd_run_all.call_args[0])


def test_pkg_install_verify_ssl_false():
    """
    test pkg.install using verify_ssl=False
    """
    ret_reg = {"Nullsoft Install System": "3.03"}
    # The 2nd time it's run, pkg.list_pkgs uses with stringify
    se_list_pkgs = [{"nsis": ["3.03"]}, {"nsis": "3.02"}]
    mock_cache_file = MagicMock(return_value="C:\\fake\\path.exe")
    with patch.object(win_pkg, "list_pkgs", side_effect=se_list_pkgs), patch.object(
        win_pkg, "_get_reg_software", return_value=ret_reg
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cmd.run_all": MagicMock(return_value={"retcode": 0}),
            "cp.cache_file": mock_cache_file,
            "cp.is_cached": MagicMock(return_value=False),
            "cp.hash_file": MagicMock(return_value={"hsum": "abc123"}),
        },
    ):
        expected = {"nsis": {"new": "3.02", "old": "3.03"}}
        result = win_pkg.install(name="nsis", version="3.02", verify_ssl=False)
        mock_cache_file.assert_called_once_with(
            "http://download.sourceforge.net/project/nsis/NSIS%203/3.02/nsis-3.02-setup.exe",
            saltenv="base",
            source_hash="abc123",
            verify_ssl=False,
            use_etag=True,
        )
        assert expected == result


def test_pkg_install_single_pkg():
    """
    test pkg.install pkg with extra_install_flags
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": MagicMock(
                return_value=[{"firebox": "3.03"}, None]
            ),
            "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
            "cmd.run_all": mock_cmd_run_all,
        },
    ):
        win_pkg.install(
            pkgs=["firebox"],
            version="3.03",
            extra_install_flags="-e True -test_flag True",
        )
        assert "-e True -test_flag True" in str(mock_cmd_run_all.call_args[0])


def test_pkg_install_log_message(caplog):
    """
    test pkg.install pkg with extra_install_flags
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": MagicMock(
                return_value=[{"firebox": "3.03"}, None]
            ),
            "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
            "cmd.run_all": mock_cmd_run_all,
        },
    ), caplog.at_level(
        logging.DEBUG
    ):
        win_pkg.install(
            pkgs=["firebox"],
            version="3.03",
            extra_install_flags="-e True -test_flag True",
        )
        assert (
            'PKG : cmd: C:\\WINDOWS\\system32\\cmd.exe /c "runme.exe" /s -e '
            "True -test_flag True"
        ).lower() in [x.lower() for x in caplog.messages]
        assert "PKG : pwd: ".lower() in [x.lower() for x in caplog.messages]
        assert "PKG : retcode: 0" in caplog.messages


def test_pkg_install_multiple_pkgs():
    """
    test pkg.install pkg with extra_install_flags
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": MagicMock(
                return_value=[{"firebox": "3.03", "got": "3.03"}, None]
            ),
            "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
            "cmd.run_all": mock_cmd_run_all,
        },
    ):
        win_pkg.install(
            pkgs=["firebox", "got"], extra_install_flags="-e True -test_flag True"
        )
        assert "-e True -test_flag True" not in str(mock_cmd_run_all.call_args[0])


def test_pkg_install_minion_error_https():
    """
    Test pkg.install when cp.cache_file encounters a minion error
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "https://repo.test.com/runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    err_msg = (
        "Error: [Errno 11001] getaddrinfo failed reading"
        " https://repo.test.com/runme.exe"
    )
    mock_none = MagicMock(return_value=None)
    mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
    mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": mock_parse,
            "cp.is_cached": mock_none,
            "cp.cache_file": mock_minion_error,
        },
    ):
        result = win_pkg.install(
            name="firebox",
            version="3.03",
        )
        expected = (
            "Failed to cache https://repo.test.com/runme.exe\nError: [Errno 11001]"
            " getaddrinfo failed reading https://repo.test.com/runme.exe"
        )

        assert result == expected


def test_pkg_install_minion_error_salt():
    """
    Test pkg.install when cp.cache_file encounters a minion error
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "salt://software/runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    err_msg = "Error: [Errno 1] failed reading salt://software/runme.exe"
    mock_none = MagicMock(return_value=None)
    mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
    mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cp.cache_file": mock_minion_error,
            "cp.is_cached": mock_none,
            "cp.hash_file": MagicMock(return_value={"hsum": "abc123"}),
            "pkg_resource.parse_targets": mock_parse,
        },
    ):
        result = win_pkg.install(
            name="firebox",
            version="3.03",
        )
        expected = (
            "Failed to cache salt://software/runme.exe\n"
            "Error: [Errno 1] failed reading salt://software/runme.exe"
        )

        assert result == expected


def test_pkg_install_minion_error_salt_cache_dir():
    """
    Test pkg.install when cp.cache_dir encounters a minion error
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "salt://software/runme.exe",
            "cache_dir": True,
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    err_msg = "Error: [Errno 1] failed reading salt://software"
    mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
    with patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cp.cache_dir": mock_minion_error,
            "cp.hash_file": MagicMock(return_value={"hsum": "abc123"}),
        },
    ):
        result = win_pkg.install(
            name="firebox",
            version="3.03",
        )
        expected = (
            "Failed to cache salt://software\n"
            "Error: [Errno 1] failed reading salt://software"
        )

        assert result == expected


def test_pkg_remove_log_message(caplog):
    """
    test pkg.remove pkg logging
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "%program.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "runme.exe",
            "uninstall_flags": "/S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
    se_list_pkgs = {"firebox": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "pkg_resource.parse_targets": MagicMock(
                return_value=[{"firebox": "3.03"}, None]
            ),
            "cp.is_cached": MagicMock(return_value="C:\\fake\\path.exe"),
            "cmd.run_all": mock_cmd_run_all,
        },
    ), caplog.at_level(
        logging.DEBUG
    ):
        win_pkg.remove(
            pkgs=["firebox"],
        )
        assert (
            'PKG : cmd: C:\\WINDOWS\\system32\\cmd.exe /c "%program.exe" /S'
        ).lower() in [x.lower() for x in caplog.messages]
        assert "PKG : pwd: ".lower() in [x.lower() for x in caplog.messages]
        assert "PKG : retcode: 0" in caplog.messages


def test_pkg_remove_minion_error_salt_cache_dir():
    """
    Test pkg.remove when cp.cache_dir encounters a minion error
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "salt://software/runme.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "salt://software/runme.exe",
            "cache_dir": True,
            "uninstall_flags": "/U /S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    err_msg = "Error: [Errno 1] failed reading salt://software"
    mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
    mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
    se_list_pkgs = {"firebox": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cp.cache_dir": mock_minion_error,
            "cp.hash_file": MagicMock(return_value={"hsum": "abc123"}),
            "pkg_resource.parse_targets": mock_parse,
        },
    ):
        result = win_pkg.remove(name="firebox")
        expected = (
            "Failed to cache salt://software\n"
            "Error: [Errno 1] failed reading salt://software"
        )

        assert result == expected


def test_pkg_remove_minion_error_salt():
    """
    Test pkg.remove when cp.cache_file encounters a minion error
    """
    ret__get_package_info = {
        "3.03": {
            "uninstaller": "salt://software/runme.exe",
            "reboot": False,
            "msiexec": False,
            "installer": "salt://software/runme.exe",
            "uninstall_flags": "/U /S",
            "locale": "en_US",
            "install_flags": "/s",
            "full_name": "Firebox 3.03 (x86 en-US)",
        }
    }

    err_msg = "Error: [Errno 1] failed reading salt://software/runme.exe"
    mock_minion_error = MagicMock(side_effect=MinionError(err_msg))
    mock_none = MagicMock(return_value=None)
    mock_parse = MagicMock(return_value=[{"firebox": "3.03"}, None])
    se_list_pkgs = {"firebox": ["3.03"]}
    with patch.object(win_pkg, "list_pkgs", return_value=se_list_pkgs), patch.object(
        salt.utils.data, "is_true", MagicMock(return_value=True)
    ), patch.object(
        win_pkg, "_get_package_info", MagicMock(return_value=ret__get_package_info)
    ), patch.dict(
        win_pkg.__salt__,
        {
            "cp.cache_file": mock_minion_error,
            "cp.hash_file": MagicMock(return_value={"hsum": "abc123"}),
            "cp.is_cached": mock_none,
            "pkg_resource.parse_targets": mock_parse,
        },
    ):
        result = win_pkg.remove(name="firebox")
        expected = (
            "Failed to cache salt://software/runme.exe\n"
            "Error: [Errno 1] failed reading salt://software/runme.exe"
        )

        assert result == expected


@pytest.mark.parametrize(
    "v1,v2,expected",
    (
        ("2.24.0", "2.23.0.windows.1", 1),
        ("2.23.0.windows.2", "2.23.0.windows.1", 1),
    ),
)
def test__reverse_cmp_pkg_versions(v1, v2, expected):
    result = win_pkg._reverse_cmp_pkg_versions(v1, v2)
    assert result == expected, "cmp({}, {}) should be {}, got {}".format(
        v1, v2, expected, result
    )


def test__repo_process_pkg_sls():
    patch_render = patch("salt.loader.render")
    patch_opts = patch.dict(win_pkg.__opts__, {"renderer": None})
    patch_compile = patch("salt.template.compile_template", return_value="junk")
    with patch_opts, patch_render as render, patch_compile as test:
        ret = win_pkg._repo_process_pkg_sls(
            filename="junk",
            short_path_name="junk",
            ret={},
            successful_verbose=False,
            saltenv="spongebob",
        )
        assert ret is False
        test.assert_called_once_with(
            "junk", render(), None, "", "", saltenv="spongebob"
        )



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import sys

import pytest

import salt.utils.win_osinfo as win_osinfo
from tests.support.unit import TestCase


@pytest.mark.skip_unless_on_windows
class WinOsInfo(TestCase):
    """
    Test cases for salt/utils/win_osinfo.py
    """

    def test_get_os_version_info(self):
        sys_info = sys.getwindowsversion()
        get_info = win_osinfo.get_os_version_info()
        self.assertEqual(sys_info.major, int(get_info["MajorVersion"]))
        self.assertEqual(sys_info.minor, int(get_info["MinorVersion"]))
        self.assertEqual(sys_info.platform, int(get_info["PlatformID"]))
        self.assertEqual(sys_info.build, int(get_info["BuildNumber"]))
        # Platform ID is the reason for this function
        # Since we can't get the actual value another way, we will just check
        # that it exists and is a number
        self.assertIn("PlatformID", get_info)
        self.assertTrue(isinstance(get_info["BuildNumber"], int))

    def test_get_join_info(self):
        join_info = win_osinfo.get_join_info()
        self.assertIn("Domain", join_info)
        self.assertIn("DomainType", join_info)
        valid_types = ["Unknown", "Unjoined", "Workgroup", "Domain"]
        self.assertIn(join_info["DomainType"], valid_types)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


@pytest.fixture(scope="module")
def run_timeout():
    if salt.utils.platform.is_windows():
        return 180
    else:
        return 30


def test_batch_run(salt_cli, run_timeout, salt_sub_minion):
    """
    Tests executing a simple batch command to help catch regressions
    """
    ret = f"Executing run on [{repr(salt_sub_minion.id)}]"
    cmd = salt_cli.run(
        "test.echo",
        "batch testing",
        "-b 50%",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )
    assert ret in cmd.stdout


def test_batch_run_number(salt_cli, salt_minion, salt_sub_minion, run_timeout):
    """
    Tests executing a simple batch command using a number division instead of
    a percentage with full batch CLI call.
    """
    ret = "Executing run on [{}, {}]".format(
        repr(salt_minion.id), repr(salt_sub_minion.id)
    )
    cmd = salt_cli.run(
        "--batch-size=2",
        "test.ping",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )
    assert ret in cmd.stdout


def test_batch_run_grains_targeting(
    grains, salt_cli, salt_minion, salt_sub_minion, run_timeout
):
    """
    Tests executing a batch command using a percentage divisor as well as grains
    targeting.
    """
    sub_min_ret = f"Executing run on [{repr(salt_sub_minion.id)}]"
    min_ret = f"Executing run on [{repr(salt_minion.id)}]"
    cmd = salt_cli.run(
        "-C",
        "-b 25%",
        "test.ping",
        minion_tgt="G@os:{} and not localhost".format(grains["os"].replace(" ", "?")),
        _timeout=run_timeout,
    )
    assert sub_min_ret in cmd.stdout
    assert min_ret in cmd.stdout


def test_batch_exit_code(salt_cli, salt_minion, salt_sub_minion, run_timeout):
    """
    Test that a failed state returns a non-zero exit code in batch mode
    """
    cmd = salt_cli.run(
        "state.single",
        "test.fail_without_changes",
        "name=test_me",
        "-b 25%",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )
    assert cmd.returncode == 2


# Test for failhard + batch. The best possible solution here was to do something like that:
# assertRaises(StopIteration)
# But it's impossible due to nature of the tests execution via fork()


def test_batch_module_stopping_after_error(
    salt_cli, salt_minion, salt_sub_minion, run_timeout
):
    """
    Test that a failed command stops the batch run
    """

    minions_list = []
    retcode = None

    # Executing salt with batch: 1 and with failhard. It should stop after the first error.
    cmd = salt_cli.run(
        "test.retcode",
        42,
        "-b 1",
        "--out=yaml",
        "--failhard",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )

    # Parsing the output. Idea is to fetch number on minions and retcode of the execution.
    # retcode var could be overwritten in case of broken failhard but number of minions check should still fail.
    for line in cmd.stdout.splitlines():
        line = line.strip()
        if line.startswith("Executing run on"):
            minions_list.append(line)
        if line.startswith("retcode"):
            retcode = int(line.split(" ")[-1])
    # We expect to have only one minion to be run
    assert 1 == len(minions_list)
    # We expect to find a retcode in the output
    assert None is not retcode
    # We expect retcode to be non-zero
    assert 0 != retcode


def test_batch_state_stopping_after_error(
    salt_cli, salt_minion, salt_sub_minion, run_timeout
):
    """
    Test that a failed state stops the batch run
    """

    minions_list = []
    retcode = None

    # Executing salt with batch: 1 and with failhard. It should stop after the first error.
    cmd = salt_cli.run(
        "state.single",
        "test.fail_without_changes",
        "name=test_me",
        "-b 1",
        "--out=yaml",
        "--failhard",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )

    # Parsing the output. Idea is to fetch number on minions and retcode of the execution.
    # retcode var could be overwritten in case of broken failhard but number of minions check should still fail.
    for line in cmd.stdout.splitlines():
        if line.startswith("Executing run on"):
            minions_list.append(line)
        if line.startswith("retcode"):
            retcode = int(line.split(" ")[-1])
    # We expect to have only one minion to be run
    assert 1 == len(minions_list)
    # We expect to find a retcode in the output
    assert None is not retcode
    # We expect retcode to be non-zero
    assert 0 != retcode


def test_batch_retcode(salt_cli, salt_minion, salt_sub_minion, run_timeout):
    """
    Ensure that the correct retcode is given in batch runs.

    See issue #60361
    """
    cmd = salt_cli.run(
        "test.retcode",
        "23",
        "-b 1",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )

    assert cmd.returncode == 23
    # TODO: Certain platforms will have a warning related to jinja. But
    # that's an issue with dependency versions that may be due to the versions
    # installed on the test images. When those issues are sorted, this can
    # simply `not cmd.stderr`.
    assert not cmd.stderr
    assert "true" in cmd.stdout


def test_multiple_modules_in_batch(salt_cli, salt_minion, salt_sub_minion, run_timeout):
    """
    Ensure that running multiple modules at the same time works in batch.

    See issue #60361
    """
    cmd = salt_cli.run(
        "test.ping,test.retcode",
        ",23",
        "-b 1",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )

    assert cmd.returncode == 23
    # TODO: Certain platforms will have a warning related to setproctitle. But
    # that's an issue with dependency versions that may be due to the versions
    # installed on the test images. When those issues are sorted, this can
    # simply `not cmd.stderr`.
    assert not cmd.stderr


def test_batch_module_stopping_failed_respond(
    salt_cli, salt_minion, salt_sub_minion, run_timeout
):
    """
    Test that minion failed to respond to job sent and stops the batch run
    """

    minions_list = []
    retcode = None
    test_data_failed = {"failed": True}

    # Executing salt with batch: 1 and with failhard. It should stop after the first error.
    cmd = salt_cli.run(
        "test.outputter",
        test_data_failed,
        "-b 1",
        "--out=yaml",
        "--failhard",
        minion_tgt="*minion*",
        _timeout=run_timeout,
    )

    # Parsing the output. Idea is to fetch number on minions and retcode of the execution, but not 'ret' key.
    # data dictionary should be overwritten, should fail regardless of failhard
    # number of minions check should still fail.
    for line in cmd.stdout.splitlines():
        line = line.strip()
        if line.startswith("Executing run on"):
            minions_list.append(line)
        if line.startswith("retcode"):
            retcode = int(line.split(" ")[-1])
        if line.startswith("failed"):
            failure = line.split(" ")[-1]
    # We expect to have only one minion to be run
    assert 2 == len(minions_list)
    # We expect to find a retcode in the output
    assert None is not retcode
    # We expect failure to be True
    assert failure == "true"



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.ipset as ipset
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    with patch.object(ipset, "_ipset_cmd", return_value="ipset"):
        yield {ipset: {}}


def test_version():
    """
    Test for Return version from ipset --version
    """
    with patch.object(ipset, "_ipset_cmd", return_value="A"):
        mock = MagicMock(return_value="A\nB\nC")
        with patch.dict(ipset.__salt__, {"cmd.run": mock}):
            assert ipset.version() == "B"


def test_new_set():
    """
    Test for Create new custom set
    """
    assert ipset.new_set() == "Error: Set Name needs to be specified"

    assert ipset.new_set("s") == "Error: Set Type needs to be specified"

    assert ipset.new_set("s", "d") == "Error: Set Type is invalid"

    assert ipset.new_set("s", "bitmap:ip") == "Error: range is a required argument"

    mock = MagicMock(return_value=False)
    with patch.dict(ipset.__salt__, {"cmd.run": mock}):
        assert ipset.new_set("s", "bitmap:ip", range="range")


def test_delete_set():
    """
    Test for Delete ipset set.
    """
    assert ipset.delete_set() == "Error: Set needs to be specified"

    with patch.object(ipset, "_ipset_cmd", return_value="A"):
        mock = MagicMock(return_value=True)
        with patch.dict(ipset.__salt__, {"cmd.run": mock}):
            assert ipset.delete_set("set", "family")


def test_rename_set():
    """
    Test for Delete ipset set.
    """
    assert ipset.rename_set() == "Error: Set needs to be specified"

    assert ipset.rename_set("s") == "Error: New name for set needs to be specified"

    with patch.object(ipset, "_find_set_type", return_value=False):
        assert ipset.rename_set("s", "d") == "Error: Set does not exist"

    with patch.object(ipset, "_find_set_type", return_value=True):
        assert ipset.rename_set("s", "d") == "Error: New Set already exists"

    with patch.object(ipset, "_find_set_type", side_effect=[True, False]):
        with patch.object(ipset, "_ipset_cmd", return_value="A"):
            mock = MagicMock(return_value=True)
            with patch.dict(ipset.__salt__, {"cmd.run": mock}):
                assert ipset.rename_set("set", "new_set")


def test_list_sets():
    """
    Test for List all ipset sets.
    """
    with patch.object(ipset, "_ipset_cmd", return_value="A"):
        mock = MagicMock(return_value="A:a")
        with patch.dict(ipset.__salt__, {"cmd.run": mock}):
            assert ipset.list_sets() == [{"A": ""}]


def test_check_set():
    """
    Test for Check that given ipset set exists.
    """
    assert ipset.check_set() == "Error: Set needs to be specified"

    with patch.object(ipset, "_find_set_info", side_effect=[False, True]):
        assert not ipset.check_set("set")
        assert ipset.check_set("set")


def test_add():
    """
    Test for Append an entry to the specified set.
    """
    assert ipset.add() == "Error: Set needs to be specified"

    assert ipset.add("set") == "Error: Entry needs to be specified"

    with patch.object(ipset, "_find_set_info", return_value=None):
        assert ipset.add("set", "entry") == "Error: Set set does not exist"

    mock = MagicMock(return_value={"Type": "type", "Header": "Header"})
    with patch.object(ipset, "_find_set_info", mock):
        assert (
            ipset.add("set", "entry", timeout=0)
            == "Error: Set set not created with timeout support"
        )

        assert (
            ipset.add("set", "entry", packets=0)
            == "Error: Set set not created with counters support"
        )

        assert (
            ipset.add("set", "entry", comment=0)
            == "Error: Set set not created with comment support"
        )

    mock = MagicMock(return_value={"Type": "bitmap:ip", "Header": "Header"})
    with patch.object(ipset, "_find_set_info", mock):
        with patch.object(ipset, "_find_set_members", return_value="entry"):
            assert (
                ipset.add("set", "entry")
                == "Warn: Entry entry already exists in set set"
            )

        with patch.object(ipset, "_find_set_members", return_value="A"):
            mock = MagicMock(return_value="")
            with patch.dict(ipset.__salt__, {"cmd.run": mock}):
                assert ipset.add("set", "entry") == "Success"

            mock = MagicMock(return_value="out")
            with patch.dict(ipset.__salt__, {"cmd.run": mock}):
                assert ipset.add("set", "entry") == "Error: out"


def test_delete():
    """
    Test for Delete an entry from the specified set.
    """
    assert ipset.delete() == "Error: Set needs to be specified"

    assert ipset.delete("s") == "Error: Entry needs to be specified"

    with patch.object(ipset, "_find_set_type", return_value=None):
        assert ipset.delete("set", "entry") == "Error: Set set does not exist"

    with patch.object(ipset, "_find_set_type", return_value=True):
        with patch.object(ipset, "_ipset_cmd", return_value="A"):
            mock = MagicMock(side_effect=["", "A"])
            with patch.dict(ipset.__salt__, {"cmd.run": mock}):
                assert ipset.delete("set", "entry") == "Success"
                assert ipset.delete("set", "entry") == "Error: A"


def test_check():
    """
    Test for Check that an entry exists in the specified set.
    """
    assert ipset.check() == "Error: Set needs to be specified"

    assert ipset.check("s") == "Error: Entry needs to be specified"

    with patch.object(ipset, "_find_set_type", return_value=None):
        assert ipset.check("set", "entry") == "Error: Set set does not exist"

    with patch.object(ipset, "_find_set_type", return_value="hash:ip"):
        with patch.object(
            ipset,
            "_find_set_members",
            side_effect=[
                "entry",
                "",
                ["192.168.0.4", "192.168.0.5"],
                ["192.168.0.3"],
                ["192.168.0.6"],
                ["192.168.0.4", "192.168.0.5"],
                ["192.168.0.3"],
                ["192.168.0.6"],
            ],
        ):
            assert ipset.check("set", "entry")
            assert not ipset.check("set", "entry")
            assert ipset.check("set", "192.168.0.4/31")
            assert not ipset.check("set", "192.168.0.4/31")
            assert not ipset.check("set", "192.168.0.4/31")
            assert ipset.check("set", "192.168.0.4-192.168.0.5")
            assert not ipset.check("set", "192.168.0.4-192.168.0.5")
            assert not ipset.check("set", "192.168.0.4-192.168.0.5")

    with patch.object(ipset, "_find_set_type", return_value="hash:net"):
        with patch.object(
            ipset,
            "_find_set_members",
            side_effect=[
                "entry",
                "",
                "192.168.0.4/31",
                "192.168.0.4/30",
                "192.168.0.4/31",
                "192.168.0.4/30",
            ],
        ):
            assert ipset.check("set", "entry")
            assert not ipset.check("set", "entry")
            assert ipset.check("set", "192.168.0.4/31")
            assert not ipset.check("set", "192.168.0.4/31")
            assert ipset.check("set", "192.168.0.4-192.168.0.5")
            assert not ipset.check("set", "192.168.0.4-192.168.0.5")


def test_test():
    """
    Test for Test if an entry is in the specified set.
    """
    assert ipset.test() == "Error: Set needs to be specified"

    assert ipset.test("s") == "Error: Entry needs to be specified"

    with patch.object(ipset, "_find_set_type", return_value=None):
        assert ipset.test("set", "entry") == "Error: Set set does not exist"

    with patch.object(ipset, "_find_set_type", return_value=True):
        mock = MagicMock(side_effect=[{"retcode": 1}, {"retcode": -1}])
        with patch.dict(ipset.__salt__, {"cmd.run_all": mock}):
            assert not ipset.test("set", "entry")
            assert ipset.test("set", "entry")


def test_flush():
    """
    Test for Flush entries in the specified set
    """
    mock = MagicMock(side_effect=["", "A"])
    with patch.dict(ipset.__salt__, {"cmd.run": mock}):
        assert ipset.flush("set")
        assert not ipset.flush("set")
